from __future__ import annotations

from typing import cast

import pytest

from curvey import Curve, Curves
from curvey.flow import CurveShorteningFlow, SingularityFreeMeanCurvatureFlow, Solver, WillmoreFlow


@pytest.fixture()
def dumbbell():
    return Curve.dumbbell(n=100)


def _assert_flow_becomes_rounder(solver: Solver):
    solver.run()
    current = cast(Curve, solver.current)
    assert current["step"] > 0
    curves = cast(Curves, solver.history)

    for c0, c1 in zip(curves[:-1], curves[1:]):
        assert c1.roundness < c0.roundness


def test_csf(dumbbell):
    solver = CurveShorteningFlow().solver(
        initial=dumbbell,
        timestep=1e-3,
        max_step=5,
    )
    _assert_flow_becomes_rounder(solver)


def test_sf_mcf(dumbbell):
    solver = SingularityFreeMeanCurvatureFlow().solver(
        initial=dumbbell,
        timestep=0.1,
        max_step=5,
    )
    _assert_flow_becomes_rounder(solver)


def test_willmore(dumbbell):
    solver = WillmoreFlow().solver(initial=dumbbell, timestep=0.01, max_step=5)

    _assert_flow_becomes_rounder(solver)



from __future__ import annotations

import pytest
from numpy import cumsum, ones, pi, sqrt, zeros
from numpy.testing import (
    assert_approx_equal,
    assert_array_almost_equal,
    assert_array_equal,
)

from curvey.curve import Curve


@pytest.fixture()
def tri():
    return Curve([[0, 0], [1, 0], [1, 1]])


@pytest.fixture()
def star():
    return Curve.star(n=6, r0=1, r1=1.5)


@pytest.fixture()
def circle():
    return Curve.circle(n=100)


def test_n(tri):
    assert tri.n == 3


def test_edge(tri):
    assert_array_almost_equal(
        tri.edge,
        [[1, 0], [0, 1], [-1, -1]],
    )


def test_metadata():
    c0 = Curve.star(n=6, r0=1, r1=1.5)
    assert not c0.data
    c1 = c0.with_data(name="mystar", max_radius=1.5)
    assert c1["name"] == "mystar"
    assert c1["max_radius"] == 1.5
    c2 = c1.scale(2)
    assert c1["name"] == "mystar"
    assert c1["max_radius"] == 1.5
    c3 = c2.with_data(max_radius=3)
    assert c3["name"] == "mystar"
    assert c3["max_radius"] == 3
    c4 = c3.drop_data("max_radius")
    assert "max_radius" not in c4.data
    assert "name" in c4.data


def test_edge_length(tri):
    assert_array_almost_equal(tri.edge_length, [1, 1, sqrt(2)])


def test_length(tri):
    assert tri.length == sum([1, 1, sqrt(2)])


def test_cum_edge_length(tri):
    assert_array_almost_equal(tri.cum_edge_length, [1, 2, 2 + sqrt(2)])
    assert_array_equal(tri.cum_edge_length, cumsum(tri.edge_length))


def test_arclength(tri):
    assert_array_almost_equal(tri.arclength, [0, 1, 2])
    assert tri.arclength[0] == 0
    assert_array_equal(tri.arclength[1:], cumsum(tri.edge_length[:-1]))


def test_closed_arclength(tri):
    assert len(tri.closed_arclength) == (tri.n + 1)
    assert_array_equal(tri.closed_arclength[:-1], tri.arclength)
    assert tri.closed_arclength[-1] == tri.length


def test_unit_edge(tri):
    assert_array_almost_equal(tri.unit_edge, [[1, 0], [0, 1], [-1 / sqrt(2), -1 / sqrt(2)]])


def test_turning_angle(tri):
    assert_array_almost_equal(tri.turning_angle, [3 * pi / 4, pi / 2, 3 * pi / 4])


@pytest.mark.parametrize(
    "tri_repeated",
    [
        # Repeated once
        [[0, 0], [0, 0], [1, 0], [1, 1]],
        [[0, 0], [1, 0], [1, 0], [1, 1]],
        [[0, 0], [1, 0], [1, 1], [1, 1]],
        # # Repeated twice
        [[0, 0], [0, 0], [0, 0], [1, 0], [1, 1]],
        [[0, 0], [1, 0], [1, 0], [1, 0], [1, 1]],
        [[0, 0], [1, 0], [1, 1], [1, 1], [1, 1]],
    ],
)
def test_drop_repeated_pts(tri, tri_repeated):
    c = Curve(tri_repeated).drop_repeated_points()
    assert c.n == 3
    assert_array_equal(tri.points, c.points)


def test_to_length(tri):
    assert_approx_equal(tri.to_length(7).length, 7)


def test_to_area(tri):
    assert_approx_equal(tri.to_area(9).area, 9)


def test_signed_area_and_orientation(tri):
    assert_approx_equal(tri.signed_area, 0.5)
    assert tri.orientation == 1
    rev = tri.reverse()
    assert_approx_equal(rev.signed_area, -0.5)
    assert rev.orientation == -1


def test_centroid(tri):
    assert_array_almost_equal(tri.centroid, tri.points.mean(axis=0))
    assert_array_almost_equal(Curve.dumbbell(n=20).centroid, [0, 0])


def test_reverse(tri):
    # [[0, 0], [1, 0], [1, 1]]
    assert_array_equal(tri.reverse(keep_first=False).points, [[1, 1], [1, 0], [0, 0]])
    assert_array_equal(tri.reverse(keep_first=True).points, [[0, 0], [1, 1], [1, 0]])


def test_subdivide(tri: Curve):
    tri2 = tri.subdivide(1)
    assert tri2.n == tri.n * 2


@pytest.mark.parametrize("kwargs", [{"n": 1}, {"min_edge_length": 0.1}])
def test_collapse_shortest_edges(kwargs):
    c = Curve([[0, 0], [0.98, 0], [1, 0], [1, 1]]).collapse_shortest_edges(**kwargs)
    assert_array_equal(c.points, [[0, 0], [1, 0], [1, 1]])


def test_is_simple():
    assert Curve([(0, 0), (2, 0), (1, 1)]).is_simple
    assert not Curve([(0, 0), (2, 0), (1, 1), (1, -1)]).is_simple


def test_edge_intersections():
    c = Curve([(0, 0), (2, 0), (1, 1)])
    assert_array_equal(c.edge_intersections(), zeros((0, 2)))
    c = Curve([(0, 0), (2, 0), (1, 1), (1, -1)])
    assert_array_equal(c.edge_intersections(), [[1.0, 0.0]])


def test_align_to(star):
    c0 = star.translate([0.1, 0.1]).rotate(0.3)

    c1 = c0.align_to(star)
    assert_array_almost_equal(c1.points, star.points)

    t = c0.align_to(star, return_transform=True)
    c1 = c0.transform(t)
    assert_array_almost_equal(c1.points, star.points)


def test_roll(star):
    star1 = star.roll(-3)
    assert_array_equal(star1.points[0], star.points[3])


def test_roll_to(star):
    star1 = star.roll(-3)
    star1 = star1.roll_to(star)
    assert_array_equal(star.points, star1.points)


def test_register_to(star):
    star1 = star.subdivide(1)

    c0 = star1.translate([0.1, 0.1]).rotate(0.3)
    c1 = c0.register_to(star)
    assert_array_almost_equal(c1.points, star1.points)

    t = c0.register_to(star, return_transform=True)
    c1 = c0.transform(t)
    assert_array_almost_equal(c1.points, star1.points)


def test_curvature(circle):
    assert_array_almost_equal(circle.curvature, ones(circle.n), decimal=3)
    assert_array_almost_equal(circle.scale(2).curvature, ones(circle.n) / 2, decimal=3)

    assert_array_almost_equal(circle.reverse().curvature, -ones(circle.n), decimal=3)
    assert_array_almost_equal(circle.reverse().scale(2).curvature, -ones(circle.n) / 2, decimal=3)


def test_optimize_edge_lengths_to(circle, star):
    star = star.interpolate_n(circle.n).to_length(circle.length)
    err0 = ((star.edge_length - circle.edge_length) ** 2).sum()
    star = star.optimize_edge_lengths_to(circle)
    err1 = ((star.edge_length - circle.edge_length) ** 2).sum()
    assert err1 < err0



"""Curve blending and shape interpolation"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable

from numpy import arctan2, eye, ndarray, newaxis, sqrt
from typing_extensions import Literal

from .curve import Curve
from .curves import Curves
from .flow import WillmoreFlow
from .util import rotation_matrix


@dataclass
class Processed:
    """Convenience class to store a curve in both its original and processed version"""

    original: Curve
    """The original untouched curve"""

    processed: Curve
    """The curve after resampling and rescaling, etc"""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(original={self.original}, processed={self.processed})"


class LinearBlending:
    """Linear vertex position interpolation

    Source and target curves must have the same number of vertices.
    """

    def __init__(self, src: Curve, tgt: Curve):
        _ = src.check_same_n_vertices(tgt)
        self.src = src
        self.tgt = tgt

    def interpolate(self, t: ndarray) -> Curves:
        r"""Interpolate curves at the requested times

        Parameters
        ----------
        t
            Vector of length `nt` with $0 \le t \le 1$ of times to interpolate at.

        Returns
        -------
        curves
            A `Curves` object with `nt` curves. The value of `t` at each point
            is stored in the curve metadata parameter 'time'.

        """
        t_ = t.reshape((-1, 1, 1))  # (nt, 1, 1)
        src_pts = self.src.points[newaxis, :, :]  # (1, n, 2)
        tgt_pts = self.tgt.points[newaxis, :, :]  # (1, n, 2)
        pts = (1 - t_) * src_pts + t_ * tgt_pts  # (nt, n, 2)
        return Curves(self.src.with_points(p).with_data(time=ti) for (p, ti) in zip(pts, t))


class CurvatureShapeBlending:
    """Curvature-based shape blending

    Based on

    [*Rapid blending of closed curves based on curvature flow.*
    Masahiro Hirano, Yoshihiro Watanabe, and Masatoshi Ishikawa. 2017.](
        https://www.sciencedirect.com/science/article/pii/S016783961730016X)

    Parameters
    ----------
    processed0
    processed1
        The processed pair storing the original curves and processed curves. Curvature flow
        interpolation requires the processed curves to have

        1. The same number of vertices
        2. The same total length
        3. Reasonably similar edge lengths between corresponding edge pairs.

    initial
        The initial curve to start targeted curvature flow from. If not supplied this
        defaults to `pair[0].processed`, but I find curvature blending to work best when
        the initial processed curve is Willmore-flowed to a circle.

    flow
        The `WillmoreFlow` object to use for curvature flows. This can be supplied if some default
        parameters need to be overridden, but is usally left `None` by default for automatic
        construction

    Examples
    --------
    ```python
    import numpy as np
    from curvey import Curve
    from curvey.blend import CurvatureShapeBlending

    n = 128
    src = Curve.circle(n=n)
    tgt = Curve.star(6, r0=1, r1=2).interpolate_n(n).scale(1.5)
    csb = CurvatureShapeBlending.preprocess(src, tgt)
    t_interp = np.linspace(0, 1, 5)
    curves = csb.interpolate(t=t_interp, stop_tol=1e-3, interp_size='area')
    curves.subplots()
    ```

    """

    def __init__(
        self,
        processed0: Processed,
        processed1: Processed,
        initial: Curve,
        flow: WillmoreFlow | None = None,
    ):
        self.processed0 = processed0
        self.processed1 = processed1
        self.initial = initial
        self.flow = flow or WillmoreFlow(realign=False)
        self.history: list[Curves | None] | None = None
        self.curvature0 = self.flow.curvature_fn(processed0.processed)
        self.curvature1 = self.flow.curvature_fn(processed1.processed)

    @property
    def processed(self) -> tuple[Curve, Curve]:
        """The pair of processed `Curve`s"""
        return self.processed0.processed, self.processed1.processed

    @property
    def original(self) -> tuple[Curve, Curve]:
        """The pair of original `Curve`s"""
        return self.processed0.original, self.processed1.original

    def _interpolate_curvature(self, t: float) -> ndarray:
        return (1 - t) * self.curvature0 + t * self.curvature1

    def _interpolate_length(self, t: float) -> float:
        c0, c1 = self.original
        return (1 - t) * c0.length + t * c1.length

    def _interpolate_area(self, t: float) -> float:
        c0, c1 = self.original
        return (1 - t) * c0.area + t * c1.area

    def _interpolate_position(self, t: float, mode: Literal["center", "centroid"]) -> ndarray:
        c0, c1 = self.original
        return (1 - t) * getattr(c0, mode) + t * getattr(c1, mode)

    def _interpolate_once(
        self, t: float, initial: Curve | None, exact_endpoints: tuple[bool, bool], **kwargs
    ) -> tuple[Curves | None, Curve]:
        k_interp = self._interpolate_curvature(t)

        flow = self.flow.with_tgt_curvature(k_interp)

        if (t in (0, 1)) and exact_endpoints[int(t)]:
            return None, self.processed[int(t)]

        initial = initial or self.initial

        solver = self._current_solver = flow.solver(
            initial=initial,
            history=self.history is not None,
            **kwargs,
        )

        solver.log("Interpolating t = {}", t)
        solver.run()
        curve = solver.current

        # Get rid of flow-specific metadata
        curve = curve.drop_data("willmore_energy", "step", "timestep")

        # Add interpolation-specific metadata
        curve = curve.with_data(
            time=t,
            src_error=sqrt(flow.energy(curve=curve, tgt_curvature=self.curvature0)),
            tgt_error=sqrt(flow.energy(curve=curve, tgt_curvature=self.curvature1)),
            interp_error=sqrt(flow.energy(curve=curve, tgt_curvature=k_interp)),
        )

        return solver.history, curve

    def interpolate(
        self,
        t: ndarray,
        path_dependent: bool = False,
        realign: bool = True,
        interp_size: Literal["length", "area"] | None = "length",
        interp_pos: Literal["center", "centroid"] | None = "center",
        reregister: bool = False,
        exact_endpoints: tuple[bool, bool] = (False, False),
        post_process: Callable[[Curve], Curve] | None = None,
        history: bool = False,
        **kwargs,
    ) -> Curves:
        """Interpolate shape at the supplied time-points

        Parameters
        ----------
        t
            Vector of interpolation parameter values in [0, 1]

        path_dependent
            If True, the resulting interpolated curve from each timepoint is used as the starting
            point for the next interpolation. Otherwise, each timepoint's flow is independently
            started from the `self.initial` curve.

        realign
            If True, the curve after each interpolation step is aligned to the preceeding step.
            This can account for changes in position and orientation. Realignment is performed
            before transforming by the interpolated size or position.

        interp_size
            If not None, scale each interpolated curve to the area or length obtained by
            interpolating the `self.pair.orig` curves' length or area.

        interp_pos
            If not None, translate the interpolated curves' center or centroid to the position
            interpolated between the `self.pair.orig` curves' center or centroid.

        reregister
            If True, perform a final post-processing step as follows: register the final
            interpolated curve against the original curve using the `Curve.register_to`
            iterative closest point algorithm. The total change in position, orientation, and
            scale is then interpolated by `t` for each interpolated curve. This generally assumes
            that 1) t=0 and t=1 are included in the interpolation points, and 2) `interp_size`
            and `interp_pos` are enabled so that the interpolated curve is close enough to the
            original curve for ICP to be successful.

        exact_endpoints
            By default, the curves at t=0 and t=1 are found by the same targeted curvature flow
            as for the interior time points. This skips that step and simply uses
            `self.pair[t].processed` if `exact_endpoints[t] == True`, for `t = 0` and `t = 1`.

        history
            If true, the history of each interpolating flow is stored in `self.history`
            as a length(t) `list[Curves]`.

        post_process
            An option function `Curve -> Curve` to apply to each interpolated curve.

        **kwargs
            Additional kwargs passed to `WillmoreFlow.solver` every interpolation step.

        Returns
        -------
        Curves
            A `Curves` of length `len(t)`.
        """
        curves = Curves()
        self.history = [] if history else None
        initial = self.initial
        raw = []  # Before interpolating size, position, etc

        for i, t_interp in enumerate(t):
            history_i, c_interp = self._interpolate_once(
                t=t_interp,
                initial=initial,
                exact_endpoints=exact_endpoints,
                **kwargs,
            )
            raw.append(c_interp)

            if i > 0 and realign:
                c_interp = c_interp.align_to(raw[-1])

            if interp_size == "length":
                c_interp = c_interp.to_length(self._interpolate_length(t_interp))
            elif interp_size == "area":
                c_interp = c_interp.to_area(self._interpolate_area(t_interp))

            if interp_pos:
                pos = self._interpolate_position(t_interp, mode=interp_pos)
                c_interp = c_interp.translate(pos)

            if post_process:
                c_interp = post_process(c_interp)

            curves.append(c_interp)

            if path_dependent:
                initial = raw[-1]

            if self.history is not None:
                self.history.append(history_i)

        if reregister:
            curves = self._reregister(curves, t=t)

        return curves

    def _reregister(self, curves: Curves, t: ndarray) -> Curves:
        # Register the final interpolated curve to the original curve
        # A 3x3 transformation matrix
        transform = curves[-1].register_to(
            target=self.processed1.original,
            allow_scale=True,
            return_transform=True,
        )

        # Extract rotation, scale, and translation from the transform matrix
        r_cos_theta, r_sin_theta = transform[:2, 0]
        theta = arctan2(r_sin_theta, r_cos_theta)
        scale_factor = sqrt(r_cos_theta**2 + r_sin_theta**2)
        offset = transform[:2, 2]

        out = Curves()
        transform = eye(3)

        for c, ti in zip(curves, t):
            transform[:2, :2] = ((1 - ti) + scale_factor * ti) * rotation_matrix(theta * ti)
            transform[:2, 2] = offset * ti
            out.append(c.transform(transform))

        return out

    @staticmethod
    def preprocess(
        c0: Curve,
        c1: Curve,
        flow: WillmoreFlow | None = None,
        circle_stop_tol=1e-3,
    ) -> CurvatureShapeBlending:
        """Preprocess a pair of curves for curvature shape blending

        Scales both curves to a common length and optimizes their edge length distributions.
        Also flows the first curve to a circle for use as the `initial` curve for flow-based
        interpolation.
        """
        _ = c0.check_same_n_vertices(c1)
        o0, o1 = c0, c1
        c0, c1 = c0.to_length(1.0), c1.to_length(1.0)
        c1 = c1.optimize_edge_lengths_to(c0)
        c0 = c0.optimize_edge_lengths_to(c1)
        p0 = Processed(original=o0, processed=c0)
        p1 = Processed(original=o1, processed=c1)

        flow = flow or WillmoreFlow()
        solver = flow.solver(initial=c1, stop_tol=circle_stop_tol, history=False)
        initial = solver.run().current

        return CurvatureShapeBlending(
            processed0=p0,
            processed1=p1,
            initial=initial,
            flow=flow,
        )



"""Interface to the [2D Shape Structure Dataset](https://2dshapesstructure.github.io/index.html)

Download the shapes zip file [here](https://2dshapesstructure.github.io/data/ShapesJSON.zip).

The dataset groups shapes by class; shapes can be loaded either by an exact name or a class name
and an index.

```python
from curvey.shape_structure_dataset import ShapeStructureDataset

try:
    dataset = ShapeStructureDataset('~/Downloads/ShapesJSON.zip')
except FileNotFoundError:
    print("Couldn't find the ShapesJSON.zip file")
else:
    print(', '.join(dataset.class_names))
    curve = dataset.load_curve('elephant-1')  # or load_curve('elephant', 0)
    curve.plot()
```
"""

from __future__ import annotations

import json
import re
from collections.abc import Iterator
from functools import cached_property
from pathlib import Path
from typing import TypedDict
from zipfile import ZipFile

from numpy import array, ndarray

from .curve import Curve


class _JsonPoint(TypedDict):
    x: float
    y: float


class _JsonTriangle(TypedDict):
    p1: int
    p2: int
    p3: int


class _JsonShape(TypedDict):
    points: list[_JsonPoint]
    triangles: list[_JsonTriangle]


class ShapeStructureDataset:
    """Interface to the 2D Shape Structure Dataset zip file

    Parameters
    ----------
    dataset
        the path to the `ShapesJSON.zip` zip file. Download it
        [here](https://2dshapesstructure.github.io/data/ShapesJSON.zip)

    Notes
    -----
    Shapes in the dataset are actually a 2d triangular mesh, not a curve. Curves are loaded without
    any processing, assuming vertices are listed in order, so they may have irregularities, such as
    repeated points or self-intersections. Shapes with topological holes like the coffee cup are
    returned as a single curve without a care in the world.
    """

    # regex to match e.g. 'Shapes/Bone-13.json'
    _shape_file_regex = re.compile(r"Shapes/([^.]+)\.json")

    def __init__(self, dataset: str | Path):
        self.dataset = ZipFile(Path(dataset).expanduser())
        self.cache: dict[str, ndarray] = {}

    def _load_json(self, name: str, idx: int | None = True) -> _JsonShape:
        name = self._canonical_name(name, idx)
        shape_bytes = self.dataset.read(f"Shapes/{name}.json")
        return json.loads(shape_bytes)

    def load_shape(
        self,
        name: str,
        idx: int | None = None,
        load_triangles: bool = True,
    ) -> tuple[ndarray, ndarray | None]:
        """A `(n_verts, 2)` array of points and a `(n_faces, 3)` array of triangles"""
        data = self._load_json(name, idx)
        pts = array([[d["x"], d["y"]] for d in data["points"]])
        if load_triangles:
            tris = array([[d["p1"], d["p2"], d["p3"]] for d in data["triangles"]])
        else:
            tris = None
        return pts, tris

    def load_points(self, name: str, idx: int | None = None) -> ndarray:
        """Load the points from the specified dataset

        Note
        ----
        Some shapes in the dataset include repeated points. This method returns point-sets
        as-is, without any further processing.

        The dataset also includes triangulations. Use method `load_shape` to load those as well.
        """
        name = self._canonical_name(name, idx)
        if name in self.cache:
            return self.cache[name]
        pts, _ = self.load_shape(name, load_triangles=False)
        self.cache[name] = pts
        return pts

    def load_curve(self, name: str, idx: int | None = None) -> Curve:
        """Construct a `Curve` from the named shape in the dataset

        Can load curves by explicit name, e.g. `dataset.load_curve('Bone-13')`,
        or a class name and an index, e.g. `dataset.load_curve('Bone', 1)`.
        Names are case-insensitive. Curves have `drop_repeated_points` applied automatically.
        """
        name = self._canonical_name(name, idx)
        pts = self.load_points(name).copy()
        return Curve(pts).drop_repeated_points().with_data(ssd_name=name)

    def _canonical_name(self, name: str, idx: int | None = None) -> str:
        if idx is None:
            return name

        return self.names_by_class[name][idx]

    @cached_property
    def all_names(self) -> set[str]:
        """Names of the shapes in the dataset"""
        return set(self._iter_all_names())

    @cached_property
    def names_by_class(self) -> dict[str, tuple[str, ...]]:
        """Shape names grouped by their class

        Returns
        -------
        classes :
            `dict[class_name] -> tuple[object_names, ...]`
        """
        from collections import defaultdict

        classes = defaultdict(list)
        special_classes = ("image", "device", "dino")

        for name in self.all_names:
            for special in special_classes:
                if name.startswith(special):
                    classes[special].append(name)
                    is_special = True
                    break
            else:
                is_special = False

            if is_special:
                continue

            class_name = name.split("-")[0].lower()
            classes[class_name].append(name)
        return {k: tuple(sorted(v)) for k, v in classes.items()}

    @cached_property
    def class_names(self) -> tuple[str, ...]:
        """Names of the unique shape classes in the dataset"""
        return tuple(sorted(self.names_by_class.keys()))

    def _iter_all_names(self) -> Iterator[str]:
        for f in self.dataset.filelist:
            if match := self._shape_file_regex.match(f.filename):
                yield match.group(1)



from __future__ import annotations

from pathlib import Path

import pytest

from curvey import Curve
from curvey.shape_structure_dataset import ShapeStructureDataset

_DATASET_FILE = Path("~/Downloads/ShapesJSON.zip").expanduser()
_DATASET_MISSING = not _DATASET_FILE.exists()
_DATASET_MISSING_REASON = "Dataset file could not be found"


def test_missing_zip_file_raises():
    with pytest.raises(FileNotFoundError):
        _ = ShapeStructureDataset("doesnt_exist.zip")


@pytest.fixture()
def dataset():
    return ShapeStructureDataset(_DATASET_FILE)


@pytest.mark.skipif(_DATASET_MISSING, reason=_DATASET_MISSING_REASON)
def test_load_curve(dataset):
    curve = dataset.load_curve("elephant-1")
    assert isinstance(curve, Curve)



from __future__ import annotations

from typing import Any, Callable, Literal, NamedTuple, Union, cast, overload

import scipy
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from numpy import arctan2, array, asarray, concatenate, cos, cross, eye, ndarray, sin, stack
from numpy.typing import ArrayLike


def angle_to_points(theta: ndarray) -> ndarray:
    return stack([cos(theta), sin(theta)], axis=1)


def rotation_matrix(theta: float) -> ndarray:
    """A 2x2 rotation matrix"""
    cos_theta, sin_theta = cos(theta), sin(theta)
    return array([[cos_theta, -sin_theta], [sin_theta, cos_theta]])


def reflection_matrix(theta: float) -> ndarray:
    """A 2x2 reflection matrix"""
    cos_2theta, sin_2theta = cos(2 * theta), sin(2 * theta)
    return array([[cos_2theta, sin_2theta], [sin_2theta, -cos_2theta]])


@overload
def align_edges(
    src_pts: ndarray,
    src_edges: ndarray,
    tgt_pts: ndarray,
    tgt_edges: ndarray,
    return_transform: Literal[True],
) -> ndarray: ...


@overload
def align_edges(
    src_pts: ndarray,
    src_edges: ndarray,
    tgt_pts: ndarray,
    tgt_edges: ndarray,
    return_transform: Literal[False],
) -> ndarray: ...


@overload
def align_edges(
    src_pts: ndarray,
    src_edges: ndarray,
    tgt_pts: ndarray,
    tgt_edges: ndarray,
    return_transform: bool,
) -> ndarray: ...


def align_edges(
    src_pts: ndarray,
    src_edges: ndarray,
    tgt_pts: ndarray,
    tgt_edges: ndarray,
    return_transform: bool = False,
) -> ndarray:
    """Calulate mean change in angle and position between two curves

    Parameters
    ----------
    src_pts
        `(n, 2)` source vertex coordinates.

    tgt_pts
        `(n, 2)` target vertex coordinates.

    src_edges
        `(n, 2)` source edge vectors.

    tgt_edges
        `(n, 2)` target edge vectors.

    return_transform
        If True, return a 3x3 transformation matrix. Otherwise, return the transformed points.

    """
    # Calculate the mean angle between source and target edge vectors
    # Because these are unnormalized, larger edges are weighted more heavily
    cos_theta = (src_edges * tgt_edges).sum()
    sin_theta = cross(src_edges, tgt_edges).sum()
    theta = arctan2(sin_theta, cos_theta)
    # NB negative sign here bc we're post-multipying, equiv. to transposing the rot matrix
    pts = src_pts @ rotation_matrix(-theta)
    # AFTER applying the rotation, check mean change in vertex position
    offset = (tgt_pts - pts).mean(axis=0, keepdims=True)

    if return_transform:
        transform = eye(3)
        transform[:2, :2] = rotation_matrix(theta)
        transform[:2, 2] = offset
        return transform

    return pts + offset


InterpType = Literal["linear", "cubic", "pchip"]


def _periodic_interp1d(x: ndarray, f: ndarray) -> Callable[[ndarray], ndarray]:
    from scipy.interpolate import interp1d

    if x[0] != 0:
        msg = "x[0] must == 0"
        raise NotImplementedError(msg)
    if len(x) != (len(f) + 1):
        msg = "len(x) must == len(f) + 1"
        raise NotImplementedError(msg)

    f = concatenate([f, f[[0]]])
    fitted = interp1d(x, f, kind="linear", assume_sorted=True, axis=0)

    def _periodic_wrapper(xi: ndarray) -> ndarray:
        return fitted(xi % x[-1])

    return _periodic_wrapper


def _periodic_pchip(x: ndarray, f: ndarray) -> Callable[[ndarray], ndarray]:
    from scipy.interpolate import PchipInterpolator

    if x[0] != 0:
        msg = "x[0] must == 0"
        raise NotImplementedError(msg)
    if len(x) != (len(f) + 1):
        msg = "len(x) must == len(f) + 1"
        raise NotImplementedError(msg)

    # let m = len(x) - 2

    # x is (x0, x1, ..., xm, xL)
    # f is (f0, f1, fm)

    x = concatenate(
        [
            # [x[-3] - x[-1]],
            [x[-2] - x[-1]],  # negative; one before the first element, i.e. (xm -xL)
            x,  # (0, x1, x2, ..., xm, xL)
            [x[-1] + x[1]],  # > L; one past the last element (f[L] = f[0])
            # [x[-1] + x[2]],
        ]
    )

    # n.b. f is one element shorter than x so need an extra pad here
    f = concatenate(
        [
            # f[[-2]],
            f[[-1]],
            f,  # f(0), f(x1), f(x2), ..., f(x_n)
            f[[0]],  # f(L)
            f[[1]],  # f(L + x1)
            # f[[2]],
        ]
    )

    pchip = PchipInterpolator(x, f, extrapolate=False)

    def _periodic_wrapper(xi: ndarray) -> ndarray:
        return pchip(xi % x[-1])

    return _periodic_wrapper


def periodic_interpolator(
    x: ndarray,
    f: ndarray,
    typ: InterpType | str = "cubic",
) -> Callable[[ndarray], ndarray]:
    """Construct a periodic interpolator of the function f(x)

    Parameters
    ----------
    x
        `(n + 1,)` array of independent variable values.

    f
        `(n,)` or `(n, ndim)` array of function values.
        `f(x[-1])` is assumed equal to `f(x[0])`.

    typ
        The type of interpolator. One of

        - 'linear' for linear interpolation via `scipy.interpolate.interp1d`
        - `cubic` for cubic spline interpolation via `scipy.interpolate.CubicSpline`
        - 'pchip`
            - for piecewise cubic hermite interpolating polynomial via
            `scipy.interpolate.PchipInterpolator`.

    Returns
    -------
    interpolator
        A function `f` that returns interpolated values.
    """

    if typ == "linear":
        return _periodic_interp1d(x, f)

    if typ == "cubic":
        return scipy.interpolate.CubicSpline(x, concatenate([f, f[[0]]]), bc_type="periodic")

    if typ == "pchip":
        return _periodic_pchip(x, f)

    msg = f"Unrecognized interpolator type {typ}"
    raise ValueError(msg)


# A way of typing the extra *args to quiver
# Might be the empty tuple, or a single element tuple
_SingleColorOrNothing = Union[tuple[()], tuple[Any]]


class _VariableColorSpec(NamedTuple):
    """Interpret a colorspec that could be either a constant color or an array of scalars

    The reason we return both variables instead of an enum is for convenience with e.g.
    matplotlibs quiver, which takes two disjoint arguments, `c` (varied) and `color` (fixed)

    """

    fixed: Any = None
    varied: ArrayLike | None = None

    @property
    def maybe_varied(self) -> tuple[()] | tuple[ArrayLike]:
        # Quiver doesn't allow named parameter for the varied color argument,
        # so we need a little splatting magic
        if self.varied is None:
            return ()

        return (self.varied,)

    @staticmethod
    def parse(
        n_data: int | None,
        supplied: Any,
        default_varied=None,
        default_fixed=None,
    ) -> _VariableColorSpec:
        if supplied is None:
            return _VariableColorSpec(fixed=default_fixed, varied=default_varied)

        if isinstance(supplied, str):  # e.g. color='black'
            return _VariableColorSpec(fixed=supplied, varied=None)

        try:
            supplied_len = len(supplied)
        except TypeError as e:
            msg = "Expected `color` to have a length"
            raise NotImplementedError(msg) from e

        if n_data == supplied_len == 3:
            # This is ambiguous and hopefully rare
            # Could raise a warning or something but whatever
            return _VariableColorSpec(fixed=None, varied=asarray(supplied))

        if n_data == supplied_len:
            return _VariableColorSpec(fixed=None, varied=asarray(supplied))

        msg = f"Expected color to be of length {n_data}, got {supplied_len}"
        raise ValueError(msg)


def _get_ax(ax: Axes | None) -> Axes:
    if ax:
        return ax

    if len(plt.get_fignums()) == 0:
        # making a new axes so don't feel bad about setting some defaults
        _fig, ax = plt.subplots()
        ax = cast(Axes, ax)
        ax.axis("equal")
        return ax

    return plt.gca()


def _rescale(v, vout: tuple[float, float] | None = None) -> ndarray | None:
    if v is None:
        return None

    if vout is None:
        return v
    v1 = plt.Normalize()(v)  # Scale to 0, 1
    return vout[0] + v1 * (vout[1] - vout[0])



"""Definitions of flow infrastructure and implementations of some common flows"""

from __future__ import annotations

import enum
import logging
from abc import abstractmethod
from copy import copy
from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    TypedDict,
    TypeVar,
    cast,
)

import numpy as np
import scipy
from numpy import clip, eye, nan, ndarray, newaxis, ones, sqrt
from typing_extensions import Self

from .curve import Curve
from .curves import Curves
from .util import InterpType

logger = logging.getLogger(__name__)


class _BraceMessage:
    """For lazy {} style log formatting"""

    def __init__(self, msg: str | None = None, *args: Any, **kwargs: Any):
        self.msg = [msg] if msg else []
        self.args = list(args)
        self.kwargs = kwargs
        self._str: str | None = None

    def __str__(self):
        if self._str is None:
            msg = "".join(self.msg)
            self._str = msg.format(*self.args, **self.kwargs)
        return self._str

    def append(self, msg: str, *args, **kwargs):
        self.msg.append(msg)
        self.args.extend(args)
        self.kwargs.update(kwargs)


TData = TypeVar("TData")
TData.__doc__ = "Flow-specific solver data type"


class AbstractFlow(Generic[TData]):
    """Abstract superclass for curve flow

    The basic contract is that `Flow` objects don't maintain any state specific
    to the solution of a flow. All state is stored in the curve metadata
    or in the `Solver.data` class, which is generic over the `TData` type.

    Two methods for subclasses to implement: `step`, which steps the curve by the
    supplied timestep, and `solver`, which constructs the auxillary `Solver` object with
    flow-specific `data: TData`.

    """

    @abstractmethod
    def solver(self, initial: Curve, **kwargs) -> Solver[TData]:
        """Construct a `Solver` to solve curve flow over time

        **kwargs are all passed to the `Solver` constructor.
        """
        ...

    @abstractmethod
    def step(self, curve: Curve, timestep: float, solver: Solver[TData]) -> Curve:
        """Step the curve by `timestep`"""
        ...

    def poststep(self, curve: Curve, solver: Solver[TData]) -> Curve:  # noqa: ARG002
        """Called after stepping the curve, but before logging it

        This is called after attaching additional curve metadata requested by curve loggers.

        Subclasses can raise `RetryStep` or `StopEarly` here if necessary, or further process the
        curve.
        """
        return curve


class RetryStep(Exception):
    """This can be raised in a custom `Solver.step_fn` to retry the current step

    Usually after adjusting the timestep or some other state.
    """


class StopEarly(Exception):
    """This can be raised in a custom `Solver.step_fn` to stop the current run

    Usually after reaching some stopping criterion.
    """


class Solver(Generic[TData]):
    """Auxillary class for solving curve `Flow`s

    Parameters
    ----------
    flow
        The `Flow` object we're solving

    initial
        The initial `Curve` to start solving from

    timestep
        For fixed timesteps

    timestep_fn
        A function `Solver -> float` that can adaptively decide a timestep on each iteration.

    history
        If true, the `Curve` after each iteration is stored in `Solver.history`, a `Curves` object.

    max_step
        Maximum number of iterations to run.

    stop_on_non_simple
        A step whose `Curve.is_simple == False` stops the run, discarding the non-simple
        curve.

    verbose
        If true, curve state information and stopping messages are printed to stdout on each
        iteration.

    log
        If true, the printed log messages as in `verbose` are saved as a list of `str`s in
        `Solver.log`

    data
        Flow specific data.

    step_fn
        A function `Solver -> Curve` that steps the curve forward at each iteration. This just
        defaults to `Solver.step`.

    """

    def __init__(
        self,
        *,
        flow: AbstractFlow,
        initial: Curve,
        data: TData,
        timestep: float | None = None,
        timestep_fn: Callable[[Solver], float] | None = None,
        history: bool = True,
        max_step: int | None = None,
        stop_on_non_simple: bool = False,
        verbose: bool = False,
        log: bool = False,
        step_fn: Callable[[Solver], Curve] | None = None,
    ):
        self.flow = flow
        self.initial = initial
        self.current = initial.with_data(time=0, step=0)
        self.previous: Curve | None = None
        self.timestep = timestep
        self.timestep_fn = timestep_fn
        self.history: Curves | None = None
        if history:
            # Note that we don't log the first curve until
            # self.run() in case some extra initialization needs to happen somewhere
            self.history = Curves([])
        self.log_history: list[str] | None = [] if log else None
        self.max_step = max_step
        self.stop_on_non_simple = stop_on_non_simple
        self.verbose = verbose

        self._stop_fns: list[Callable[[Solver], bool]] = []
        self._curve_loggers: dict[str, Callable[[Curve], Any]] = {}
        self.data = data
        self.step_fn = step_fn

    def __repr__(self) -> str:
        solver_name = self.__class__.__name__
        flow_name = self.flow.__class__.__name__
        return f"{solver_name}(flow={flow_name}, current={self.current})"

    def log(self, msg: str, *args, **kwargs):
        """Log a message

        This is always sent to the module `logging.logger` at debug level.
        If `self.verbose` is true, it's also printed to stdout.
        If `self.log` is true, it's saved as a str in self.log_history
        """
        self._log(_BraceMessage(msg, *args, **kwargs))

    def _log(self, bm: _BraceMessage):
        logger.debug(bm)
        if self.verbose:
            pass
        if self.log_history is not None:
            self.log_history.append(str(bm))

    def _log_state(self, msg: str):
        m = _BraceMessage(msg=msg)

        for k, v in self.current.data.items():
            if k != "step":
                m.append(", {} = {}", k, v)

        self._log(m)

    def _log_step(self, c1: Curve):
        m = _BraceMessage("Step {}", c1["step"])
        c0 = self.current
        ks = (c0.data.keys() & c1.data.keys()) - {"step"}
        for k in ks:
            m.append(", {}: {} => {}", k, c0[k], c1[k])
        self._log(m)

    def _stop_fn(self) -> bool:
        """Returns True if run should stop"""
        if self.max_step is not None and self.current["step"] == self.max_step:
            self.log("Stopping at max step {}", self.current["step"])
            return True

        for fn in self._stop_fns:
            if fn(self):
                self.log("Stopping due to stop fn {}", fn)
                return True

        return False

    def add_stop_fn(self, fn: Callable[[Solver], bool]) -> Self:
        """Add a custom stop function. The run is stopped early if `fn(curve)` returns True

        Returns
        -------
        self

        """
        self._stop_fns.append(fn)
        return self

    def add_curve_loggers(self, **kwargs: Callable[[Curve], Any]) -> Self:
        """Log additional information as curve metadata

        e.g. `solver.add_curve_loggers(foo=foo_fn, bar=bar_fn) will store the results of
        the function calls `foo_fn(curve)` and `bar_fn(curve)` in the curve metadata 'foo' and 'bar'
        properties.

        Returns
        -------
        self

        """
        self._curve_loggers.update(kwargs)
        return self

    def stop_on_param_limits(
        self,
        param: str,
        min_val=None,
        max_val=None,
        param_fn: Callable[[Curve], Any] | None = None,
    ) -> Self:
        """Add a custom stop function based on a parameter value

        Parameters
        ----------
        param
            The name of the parameter. This is usually a curve metadata object, e.g. one logged
            via `Solver.add_curve_loggers`. If `param_fn` is supplied, `param` is only used as the
            parameter name for logging purposes.

        min_val
            The run is stopped if the parameter value < `min_val`. `None` means no lower limit.

        max_val
            The run is stopped if the parameter value > `max_val`. `None` means no upper limit.

        param_fn
            An optional function `Curve -> value`; if supplied `param` doesn't need to be available
            as `Curve` metadata.

        Returns
        -------
        self

        """
        if param_fn is None:

            def param_fn(c: Curve) -> Any:
                return c[param]

        def param_limits_stop_fn(solver: Solver) -> bool:
            val = param_fn(solver.current)

            if min_val is not None and val < min_val:
                solver.log("Parameter {} value {} < min value {}, stopping", param, val, min_val)
                return True

            if max_val is not None and val > max_val:
                solver.log("Parameter {} value {} > max value {}, stopping", param, val, max_val)
                return True

            return False

        self.add_stop_fn(param_limits_stop_fn)
        return self

    def run(self):
        """Solve the flow by stepping the curve through time

        If no stop criterion are specified by `max_step`, `add_stop_fn`, or
        `stop_on_param_limits`, this might run forever.

        Returns
        -------
        self

        """
        self.current = self.attach_metadata(self.initial, time=0, step=0)
        if self.history is not None:
            self.history.append(self.current)

        self._log_state("Initial state")
        step_fn = self.step_fn or Solver.step

        while not self._stop_fn():
            try:
                next_curve = step_fn(self)
                next_curve = self.flow.poststep(curve=next_curve, solver=self)
            except RetryStep:
                continue
            except StopEarly:
                break

            if self.stop_on_non_simple and not next_curve.is_simple:
                self.log("Curve is non-simple, stopping")
                break

            self._log_step(next_curve)
            self.previous = self.current
            self.current = next_curve
            if self.history is not None:
                self.history.append(next_curve)

        self._log_state("Final state")
        return self

    def step(self) -> Curve:
        """Call `Flow.step` with the current state and timestep"""

        if self.timestep_fn is not None:
            timestep = self.timestep_fn(self)
        elif self.timestep is not None:
            timestep = self.timestep
        else:
            msg = "Neither of `timestep` or `timestep_fn` were provided."
            raise ValueError(msg)

        # The `Flow` does the actual work here
        curve = self.flow.step(
            curve=self.current,
            timestep=timestep,
            solver=self,
        )

        return self.attach_metadata(
            curve=curve,
            time=self.current["time"] + timestep,
            timestep=timestep,
        )

    def attach_metadata(
        self, curve: Curve, time: float, step: int | None = None, **kwargs
    ) -> Curve:
        """Store requested metadata on the curve

        Parameters
        ----------
        curve
            The curve after the most recent step.

        time
            The time of the curve in the solution.

        step
            Which step this curve belongs to. This is almost always left None; it defaults
            to `solver.current['step'] + 1`.

        **kwargs
            Additional metadata to store as key=value pairs.

        Returns
        -------
        curve
            The curve with metadata attached.
        """
        if step is None:
            step = self.current["step"] + 1

        params = dict(time=time, step=step, **kwargs)
        for k, fn in self._curve_loggers.items():
            params[k] = fn(curve)
        return curve.with_data(**params)


class AbstractCurvatureFlow(AbstractFlow[TData], Generic[TData]):
    """Abstract superclass for curvature flows

    Parameters
    ----------
    curvature_fn
        A function `Curve -> ndarray` that returns the signed curvature values at each vertex.

    rescale
        If this is 'length' or 'area', the recurve length/area is rescaled to the original's
        length or area, preventing the usual curvature flow shrinkage.
    """

    def __init__(
        self,
        curvature_fn: Callable[[Curve], ndarray] | None = None,
        rescale: Literal["length", "area"] | None = None,
    ):
        super().__init__()
        self.curvature_fn = curvature_fn or self.default_curvature_fn
        self.rescale = rescale

    @abstractmethod
    def step(self, curve: Curve, timestep: float, solver: Solver[TData]) -> Curve: ...

    @abstractmethod
    def solver(self, initial: Curve, **kwargs) -> Solver[TData]: ...

    def _postprocess(self, curve: Curve, solver: Solver[TData]) -> Curve:
        if self.rescale == "area":
            curve = curve.scale(sqrt(solver.initial.area / curve.area))
        elif self.rescale == "length":
            curve = curve.scale(solver.initial.length / curve.length)

        return curve

    @staticmethod
    def default_curvature_fn(curve: Curve) -> ndarray:
        """Simply return the value of the `Curve.curvature` property"""
        return curve.curvature


class _CsfData(TypedDict):
    orig_thresh: float


class CurveShorteningFlow(AbstractCurvatureFlow[_CsfData]):
    r"""Basic curve shortening flow

    At each iteration, vertices coordinates are moved by $\Delta t \kappa_i N_i$, for
    timestep $\Delta t$ and vertex curvatures $\kappa_i$ and normal $N_i$.

    Parameters
    ----------
    resample_mode
        Type of interpolation to use when resampling, one of ('linear', 'cubic', 'pchip').

    **kwargs
        Remaining kwargs are passed to the `AbstractCurvatureFlow` constructor.
    """

    def __init__(
        self,
        resample_mode: InterpType | None = "cubic",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.resample_mode = resample_mode

    def solver(self, initial: Curve, **kwargs) -> Solver[_CsfData]:
        """Construct a `CurveShorteningFlow` `Solver`

        **kwargs are all passed to the `Solver` constructor.
        """
        data = _CsfData(orig_thresh=initial.edge_length.mean())
        return Solver(flow=self, initial=initial, data=data, **kwargs)

    def step(self, curve: Curve, timestep: float, solver: Solver[_CsfData]) -> Curve:
        curve = curve.translate(timestep * self.curvature_fn(curve)[:, newaxis] * curve.normal)
        curve = super()._postprocess(curve=curve, solver=solver)

        if self.resample_mode:
            curve = curve.interpolate_thresh(
                thresh=solver.data["orig_thresh"],
                typ=self.resample_mode,
            )

        return curve


class _Sentinel(enum.Enum):
    DEFAULT = object()


class _WillmoreFlowData(TypedDict):
    stop_on_energy_increase: bool


class WillmoreFlow(AbstractCurvatureFlow[_WillmoreFlowData]):
    r"""Willmore Flow

    As explained in [*Robust Fairing via Conformal Curvature Flow.* Keenan Crane, Ulrich Pinkall,
    and Peter Schröder. 2014.](
    https://www.cs.cmu.edu/~kmcrane/Projects/ConformalWillmoreFlow/paper.pdf)

    Parameters
    ----------
    filter_width
    filter_shape
        The $\theta$ and $k$ parameters in Crane §4. These filter the curvature flow direction and
        can be used to prioritize high or low frequency smoothing.

    constrain
        Whether to apply the closed curve constraints on the curvature flow direction at each
        timestep. See method `WillmoreFlow.constrain_flow` for more details.

    solve_vertices
        Whether to distribute length discretization errors.
        See method `Curve.with_curvatures` for more details.

    realign
        Whether to realign the curve at each timestep to the preceeding one. Because flipping
        back and forth between extrinsic and intrinsic representations loses rotation and
        translation information, this helps visually align the curve at each step, but may be
        an unnecessary computation each iteration if alignment isn't important. See method
        `Curve.with_curvatures` for more details.

    tgt_curvature
        Vector of target vertex curvatures to flow towards.

    """

    def __init__(
        self,
        constrain: bool = True,
        filter_width: float | None = None,
        filter_shape: int | None = None,
        solve_vertices: bool = True,
        realign: bool = True,
        tgt_curvature: ndarray | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.constrain = constrain
        self.filter_width = filter_width
        self.filter_shape = filter_shape
        _filter_params_specified = {filter_shape is not None, filter_width is not None}
        if len(_filter_params_specified) != 1:
            msg = (
                "Both `filter_width` and `filter_shape` must be specified to filter curvature flow"
            )
            raise ValueError(msg)
        self._do_filter: bool = _filter_params_specified.pop()

        self.solve_vertices = solve_vertices
        self.realign = realign
        self.tgt_curvature = tgt_curvature

    @staticmethod
    def constrain_flow(curve: Curve, dk: ndarray) -> ndarray:
        """Constrain curvature flow as per Crane §5

        Constraints are

        1. end points must meet: $f(0) = f(L)$
        2. tangents must agree at endpoints: $T(0) = T(L)$.

        Parameters
        ----------
        curve
            The curve to constrain flow for.

        dk
            A `n_vertices` length vector indicating the curvature flow.

        Returns
        -------
        dk_constrained
            the curvature flow direction after applying the constraints.

        """
        mass = curve.dual_edge_length

        def inner_product(f: ndarray, g: ndarray) -> float:
            """The L2 inner product ⟨⟨F G⟩⟩"""
            return (f * mass * g).sum()

        def proj(f: ndarray, g: ndarray) -> ndarray:
            """Projection of f onto g"""
            return inner_product(f, g) / inner_product(g, g) * g

        # Construct orthogonal constraint basis (Crane §4, the `c_i` terms) via Gram-Schmidt
        x, y = curve.points.T
        c0 = ones(curve.n)
        c1 = x - proj(x, c0)
        c2 = y - proj(y, c1) - proj(y, c0)

        # Subtract flow along the constraint basis
        return dk - proj(dk, c0) - proj(dk, c1) - proj(dk, c2)

    def filter_flow_direction(self, curve: Curve, dk: ndarray) -> ndarray:
        """Filter curvature flow gradient"""
        sigma, order = self.filter_width, self.filter_shape
        if sigma is None:
            msg = "filter_width is None"
            raise ValueError(msg)
        if order is None:
            msg = "filter_shape is None"
            raise ValueError(msg)

        n = curve.n

        # Square matrix `a` here is the term `id - σ∆^k` in Crane §4
        # `filtered` is inv(a) @ dk
        if order == 0:
            # can't be a sparse array
            a = eye(n, n) - sigma * ones((n, n))
            filtered = np.linalg.solve(a, dk)
        else:
            # lb = curve.laplacian
            lb = scipy.sparse.diags_array(-1 / curve.dual_edge_length) @ curve.laplacian
            a = scipy.sparse.eye(n, n) - sigma * (lb**order)
            filtered = scipy.sparse.linalg.spsolve(a, dk)

        return dk - filtered  # v ← v - inv(a)v

    def solver(
        self,
        initial: Curve,
        stop_tol: float | None = None,
        stop_on_energy_increase: bool = False,
        **kwargs,
    ) -> Solver[_WillmoreFlowData]:
        """Construct a `Solver` for the flow

        Parameters
        ----------
        initial
            The initial `Curve`.

        stop_tol
            Optional stopping tolerance. See `WillmoreFlow.stop_on_gradient_tolerance`
            for more details.

        stop_on_energy_increase
            Stop the first time energy is increased. The step with increased energy is discarded.

        **kwargs
            Remaining kwargs passed to the `Solver` constructor.

        Notes
        -----
        If neither `timestep` nor `timestep_fn` are supplied to the solver, sets the solver
        `timestep_fn` to `self.autotimestep_fn` for adaptive timestep selection. When
        `tgt_curvatures` is None, it's probably safe to just use a reasonably large timestep < 1,
        but an adaptive timestep seems to be safer for targeted curvature flow.
        """

        data = _WillmoreFlowData(stop_on_energy_increase=stop_on_energy_increase)
        solver = Solver(flow=self, initial=initial, data=data, **kwargs)
        solver.add_curve_loggers(willmore_energy=self.energy)

        if solver.timestep is solver.timestep_fn is None:
            solver.timestep_fn = self.autotimestep_fn()

        if stop_tol is not None:
            solver.add_stop_fn(self.stop_on_gradient_tolerance(stop_tol))

        return solver

    def step(self, curve: Curve, timestep: float, solver: Solver[_WillmoreFlowData]) -> Curve:
        """Step the curve along its Willmore energy gradient"""
        k0 = self.curvature_fn(curve)

        # Calculate curvature gradient, i.e. the derivative of E(k) = ||k||^2
        dk = -2 * k0 if self.tgt_curvature is None else -2 * (k0 - self.tgt_curvature)

        if self._do_filter:
            dk = self.filter_flow_direction(curve, dk)

        if self.constrain:
            dk = self.constrain_flow(curve, dk)

        k1 = k0 + timestep * dk
        curve = curve.with_curvature(
            curvature=k1,
            solve_vertices=self.solve_vertices,
            realign=self.realign,
        )
        return self._postprocess(curve, solver=solver)

    def poststep(self, curve: Curve, solver: Solver[_WillmoreFlowData]) -> Curve:
        if solver.data["stop_on_energy_increase"]:
            e1 = curve["willmore_energy"]
            e0 = solver.current["willmore_energy"]
            if e1 > e0:
                solver.log("Willmore energy increased {} => {}, stopping", e0, e1)
                raise StopEarly()

        return curve

    def energy(
        self,
        curve: Curve,
        tgt_curvature: ndarray | _Sentinel | None = _Sentinel.DEFAULT,
    ) -> float:
        r"""Calculate curve energy

        By default uses `self.tgt_curvature`, but can be overridden with the supplied
        `tgt_curvature`.

        If `tgt_curvature` is None, calculates the Willmore energy

        $$
            E(c) = \sum_i^n \kappa_i^2 l_i$
        $$

        for vertex curvatures $\kappa_i$ and dual edge lengths $l_i$.

        If `tgt_curvature` is not None, calculates

        $$
            E(c) = \sum_i^n ( \kappa_i - \hat \kappa_i)^2 l_i
        $$

        for target vertex curvatures $\hat \kappa_i$.

        """
        if tgt_curvature is _Sentinel.DEFAULT:
            tgt_curvature = self.tgt_curvature

        if tgt_curvature is None:
            return (self.curvature_fn(curve) ** 2 * curve.dual_edge_length).sum()

        dk = self.curvature_fn(curve) - tgt_curvature
        return (dk**2 * curve.dual_edge_length).sum()

    @staticmethod
    def autotimestep_fn(
        min_step: float | None = 1e-5,
        max_step: float | None = 0.9,
    ) -> Callable[[Solver], float]:
        r"""Construct an adaptive timestep function

        For curve $c$, calculates the timestep as $1 / \sqrt {E(c)}$, for energy $E$, defined
        in `WillmoreFlow.energy`. This value is then clamped to `min_step` and `max_step`,
        if supplied.

        Parameters
        ----------
        min_step
            Minimum timestep.

        max_step
            Maximum timestep

        Returns
        -------
        timestep_fn
            A function `Solver -> timestep`.
        """

        def timestep_fn(solver: Solver) -> float:
            # NB WillmoreFlow.solver adds an energy logger
            e = solver.current["willmore_energy"]
            if e == 0:
                # Doesn't matter what the stepsize is
                dt = solver.current.data.get("timestep", nan)
                return cast(float, np.nan_to_num(dt))

            return float(clip(1 / sqrt(e), min_step, max_step))

        return timestep_fn

    @staticmethod
    def stop_on_gradient_tolerance(tol: float) -> Callable[[Solver], bool]:
        """Construct a `Solver` stopping function for the supplied tolerance"""

        def stop_fn(solver: Solver):
            if solver.previous is None:
                return False

            e1 = solver.current["willmore_energy"]
            e0 = solver.previous["willmore_energy"]
            dt = solver.current["timestep"]
            gradient = abs(e1 - e0) / dt
            if gradient < tol:
                solver.log("Willmore energy gradient {} < tol {}, stopping", gradient, tol)
                return True

            return False

        return stop_fn

    def with_tgt_curvature(self, tgt_curvature: ndarray) -> WillmoreFlow:
        """Replace the target curvatures"""
        out = copy(self)
        out.tgt_curvature = tgt_curvature
        return out


class _SfmcfData(TypedDict):
    stiffness: scipy.sparse.sparray


class SingularityFreeMeanCurvatureFlow(AbstractCurvatureFlow):
    """Singularity free mean curvature flow

    As defined in

    [*Can Mean-Curvature Flow Be Made Non-Singular?* Michael Kazhdan, Jake Solomon, and Mirela Ben-Chen.
    2012.](https://arxiv.org/abs/1203.6819)

    That paper suggests this shouldn't really be necessary in the planar curve case, as curves in
    the continuous case can't form singularities anyway, but it does seem to be much more
    numerically stable than the traditional approach, and doesn't require resampling the curve.

    See also the explanation in

    [*Mean Curvature Flow and Applications*. Maria Eduarda Duarte and Leonardo Sacht. 2017.](
    https://sibgrapi.sid.inpe.br/col/sid.inpe.br/sibgrapi/2017/09.04.18.39/doc/Mean%20Curvature%20Flow%20and%20Applications.pdf)

    Parameters
    ----------
    **kwargs
        All kwargs passed to `AbstractCurvatureFlow` constructor.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def solver(self, initial: Curve, **kwargs) -> Solver[_SfmcfData]:
        """Construct a solver for this flow

        All **kwargs are passed to the `Solver` constructor.
        """
        stiffness = scipy.sparse.diags(-1 / initial.dual_edge_length) @ initial.laplacian
        data = _SfmcfData(stiffness=stiffness)
        return Solver(flow=self, initial=initial, data=data, **kwargs)

    def step(self, curve: Curve, timestep: float, solver: Solver[_SfmcfData]) -> Curve:
        inv_mass = scipy.sparse.diags(1 / curve.dual_edge_length)  # The $D_t^-1$ matrix
        stiffness = solver.data["stiffness"]  # The stiffness $L_0$
        # noinspection PyTypeChecker
        pts = scipy.sparse.linalg.spsolve(inv_mass - timestep * stiffness, inv_mass @ curve.points)
        curve = solver.initial.with_points(pts)
        return super()._postprocess(curve=curve, solver=solver)



from __future__ import annotations

import pytest
from numpy.testing import assert_array_equal

from curvey import Curve, Curves
from curvey.curves import _SubplotsBuilder


@pytest.fixture()
def curve() -> Curve:
    return Curve.circle(n=3)


def curves(n: int) -> Curves:
    c = Curve.circle(n=3)
    return Curves([c for _ in range(n)])


def test_empty():
    cs = Curves()
    assert cs.n == 0

    cs = Curves([])
    assert cs.n == 0


def test_from_list(curve):
    n = 3
    cs = Curves([curve for _ in range(n)])
    assert cs.n == n


def test_from_generator(curve):
    n = 3
    cs = Curves(curve for _ in range(n))
    assert cs.n == n


def test_subplots_builder_get_dims():
    assert _SubplotsBuilder.get_dims(20, 1, 5) == (1, 5)


def test_subplots_builder_from_dims():
    cs = curves(n=20)
    b = _SubplotsBuilder.from_dims(cs, nr=1, nc=5)
    assert b.n_axs == 5
    assert b.n_plots == 5
    assert_array_equal(b.plot_idxs, [0, 4, 8, 12, 16])



from __future__ import annotations

import numpy as np
import pytest
from numpy.testing import assert_array_equal
from shapely import hausdorff_distance

from curvey import Curve
from curvey.blend import CurvatureShapeBlending, LinearBlending


@pytest.fixture()
def endpoints() -> tuple[Curve, Curve]:
    n = 128
    src = Curve.circle(n=n)
    tgt = Curve.star(6, r0=1, r1=2).interpolate_n(n).scale(1.5)
    return src, tgt


def test_linear_blending(endpoints):
    lb = LinearBlending(*endpoints)
    t = np.linspace(0, 1, 5)
    curves = lb.interpolate(t=t)
    assert curves.n == len(t)

    assert_array_equal(curves[0].points, endpoints[0].points)
    assert_array_equal(curves[-1].points, endpoints[1].points)


def test_curvature_shape_blending(endpoints):
    csb = CurvatureShapeBlending.preprocess(*endpoints)
    t = np.linspace(0, 1, 5)
    curves = csb.interpolate(t=t, stop_tol=1e-3, interp_size="area")
    assert curves.n == len(t)

    o0, o1 = (c.to_shapely("polygon") for c in endpoints)
    i0, i1 = (curves[i].to_shapely("polygon") for i in (0, -1))

    assert hausdorff_distance(o0, i0) < 0.1
    assert hausdorff_distance(o1, i1) < 0.1



"""Sequences of curves"""

from __future__ import annotations

import collections
from collections.abc import Iterable, Iterator, Sequence
from dataclasses import dataclass
from functools import cached_property
from typing import (
    Any,
    Callable,
    Literal,
    Union,
    cast,
    overload,
)

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from numpy import arange, array, asanyarray, asarray, ceil, nan, ndarray, sqrt
from numpy.typing import ArrayLike

from .curve import Curve
from .util import _get_ax

NamedMetadata = Union[tuple[str, Sequence], tuple[str, Callable[[Curve], Any]]]
BareMetadata = Union["str", Sequence, Callable[[Curve], Any]]
Metadata = Union[NamedMetadata, BareMetadata]


class Curves:
    """A container of `Curve`s

    The constructor accepts an iterator of `Curve`s:
    ```python
    from curvey import Curve, Curves

    curves = Curves(Curve.circle(n=20, r=r) for r in (1, 2, 3))
    ```

    `Curves` can be iterated over:
    ```python
    for c in curves:
        ...
    ```

    `Curves` can be indexed with:

    - an int: `curves[i]` returns the `Curve` at the `i`ith position

    - a slice or numpy index-like array: `curves[::2]` returns a new `Curves` containing every
        second curve

    - a string: `curves['param']` returns an array of the curve metadata or `Curve` attribute
        associated with each `Curve` in the `Curves`

    - a function: `curves[fn]` is equivalent to `array([fn(c) for c in curves])`

    Use `Curves.subplots` or `Curves.superimposed` to plot every curve in the `Curves` at once.

    Use `Curves.plot` to plot `Curve` metadata values against eachother.

    Parameters
    ----------
    curves
        A iterable of `Curves`. Defaults to an empty sequence if not supplied.
    """

    def __init__(self, curves: Iterable[Curve] | None = None):
        self.curves: list[Curve] = []
        """The `Curve`s contained in this `Curves`"""

        if curves is not None:
            self.curves.extend(curves)

    @property
    def n(self) -> int:
        """Number of curves in the sequence"""
        return len(self.curves)

    def __len__(self) -> int:
        """Number of curves in the sequence"""
        return len(self.curves)

    def __add__(self, other: Curves | list[Curve]) -> Curves:
        """Concatenate two `Curves` to form a new sequence"""
        if isinstance(other, Curves):
            return Curves(self.curves + other.curves)

        # Just hope this raises a reasonable error if it fails
        return Curves(self.curves + other)

    def append(self, curve: Curve):
        """Add a curve to the end of the sequence"""
        self.curves.append(curve)

    def keys(self, mode: Literal["intersection", "union"] = "union") -> set[str]:
        """Unique curve metadata parameter names

        Parameters
        ----------
        mode
            If 'union', return keys that are present on any `Curve` in the `Curves`.
            If 'intersection', return only keys that are present on all `Curve`s in the `Curves`.
        """
        if self.n == 0:
            return set()

        keys = (set(c.data.keys()) for c in self)
        if mode == "intersection":
            return set.intersection(*keys)

        if mode == "union":
            return set.union(*keys)

        raise ValueError(mode)

    def __repr__(self) -> str:
        if keys := self.keys():
            data = ", ".join(k for k in sorted(keys))
            # Gross triple nested curlies here, so it formats like `Curves(n=3, data={foo, bar})`
            return f"Curves(n={self.n}, data={{{data}}})"

        return f"Curves(n={self.n})"

    def __iter__(self) -> Iterator[Curve]:
        """Iterate over curves in the sequence

        The index of the curve is stored in the 'idx' metadata parameter.
        Use `Curves.iter_curves` to supply a custom name for index, if necessary.
        """
        return self.iter_curves()

    def iter_curves(self, idx: str = "idx") -> Iterator[Curve]:
        """Iterate over curves in the sequence

        The index of the curve is stored in the `idx` metadata parameter. This might be useful for
        tracking the original index in a subset.

        ```python
        from curvey import Curves

        orig = Curves(Curve.circle(3) for _ in range(6))
        every_other = Curves(orig.iter_curves('orig_idx'))[::2]
        every_other['orig_idx']
        ```
        """
        for i, c in enumerate(self.curves):
            yield c.with_data(**{idx: i})

    def get_named_data(self, data: BareMetadata | NamedMetadata, **kwargs) -> tuple[str, ndarray]:
        """Get curve metadata (name, values) pairs

        If `data` is just a name, return (name, values).
        If `data` is something that can reasonably interpreted as `values`,
        try to figure out a reasonable name for them.
        """
        if isinstance(data, str):
            return data, self.get_data(data, **kwargs)

        if isinstance(data, tuple) and len(data) == 2:
            name = data[0]
            if not isinstance(name, str):
                msg = f"Expected metadata name to be str, got {name}"
                raise TypeError(msg)
            return name, self.get_data(data[1], **kwargs)

        if callable(data):
            name = data.__name__ if hasattr(data, "__name__") else str(data)
            return name, self.get_data(data, **kwargs)

        name = type(data).__name__
        return name, self.get_data(data, **kwargs)

    def get_data(
        self,
        data: BareMetadata,
        default: Any = nan,
        allow_default: bool = True,
        allow_property: bool = True,
    ) -> ndarray:
        """Concatenate curve metadata into an array of length `n`

        Parameters
        ----------
        data
            One of:

            - Name of the property stored in curve metadata
            - Name of a `Curve` attribute, if `allow_property` is true
            - A function `Curve -> value`
            - An length `n` array or list of values

        allow_property
            If true, `data` may be the name of a Curve attribute, such as 'area' or 'length'

        allow_default
            If true, and the requested data is only available on a subset of curves, return
            `default` for those curves.

        default
            The default value if named parameter `data` is not present in a curve's metadata.
            If not supplied, all curves in the collection must have that metadata parameter,
            otherwise a `KeyError` is raised.

        """
        if isinstance(data, str):
            if data in self.keys("union"):
                if allow_default:
                    return array([c.data.get(data, default) for c in self])

                if data in self.keys("intersection"):
                    msg = f"Metadata '{data}' is only present on some curves"
                    raise KeyError(msg)

                return array([c[data] for c in self])

            if hasattr(Curve, data):
                if allow_property:
                    return array([getattr(c, data) for c in self])

                msg = (
                    f"Metadata '{data}' not found. "
                    "it's a Curve property but `allow_property` is False"
                )
                raise KeyError(msg)

            msg = f"Metadata '{data}' not found"
            raise KeyError(msg)

        if callable(data):
            return array([data(c) for c in self])

        if isinstance(data, Sequence):
            if (n := len(data)) != self.n:
                msg = f"Expected a sequence of length self.n = {self.n}, got {n}"
                raise ValueError(msg)
            return asarray(data)

        msg = f"Unrecognized data type {type(data)}"  # type: ignore [unreachable]
        raise TypeError(msg)

    @overload
    def __getitem__(self, idx: int) -> Curve: ...

    @overload
    def __getitem__(self, idx: slice) -> Curves: ...

    @overload
    def __getitem__(self, idx: str | collections.abc.Callable) -> ndarray: ...

    def __getitem__(
        self,
        idx: str | int | slice | ArrayLike | collections.abc.Callable,
    ) -> Curve | Curves | ndarray:
        """Convenience method to index the sequence

        `Curves[int]` returns the curve stored at that index.
        `Curves[str]` returns a `ndarray` of `n` metadata values.
        `Curves[fn]` for `fn: Callable[[Curve], Any]` returns a `ndarray` of the values
        of that function called on the `n` curves in the sequence.

        Otherwise, `Curves[idx]` returns a new `Curves` for that index,
        obeying list slicing and numpy smart indexing behavior. E.g. `sequence[::3]` returns
        a new curve sequence for every third curve in the original sequence.

        """
        if isinstance(idx, str):
            return self.get_data(idx)

        if callable(idx):
            return self.get_data(idx)

        if isinstance(idx, (int, np.integer)):
            # Recast to int here so indexing with np.int works as expected
            return self.curves[int(idx)]

        # noinspection PyUnresolvedReferences
        idx = arange(self.n)[idx]  # type: ignore [index]
        all_curves = self.curves
        curves = [all_curves[int(i)] for i in idx]
        return Curves(curves=curves)

    def plot(
        self,
        y: Metadata,
        x: Metadata | None = None,
        ax: Axes | None = None,
        label_axes: bool | None = None,
        label: str | bool | None = True,
        **kwargs,
    ):
        """Plot metadata values against each other.

        By default, the independent variable `y` is 'time', if it's present.

        Parameters
        ----------
        y
            The name of the parameter to plot on the y-axis.
            Can also be a function `Curve` -> float, or a `tuple(name, function)`,
            Or a `tuple(name, array)`

        x
            The name of the parameter to plot on the x-axis, or alternative type as described for
            `y`. If not supplied, defaults to `time` if it's present in the curve metadata,
            or index otherwise.

        label_axes
            If true, set x and y labels. Defaults to true if a new axes is created.

        label
            Name to label the plot with, for use in matplotlib legend. Defaults to the name
            of the `y` parameter.

        ax
            The matplotlib axes to plot in. Defaults to the current axes.

        **kwargs
            Remaining kwargs are passed to `matplotlib.pyplot.plot`

        """
        if ax is None:
            ax = plt.gca()
            if label_axes is None:
                label_axes = True

        yname, ydata = self.get_named_data(y)

        if x is None:
            keys = self.keys(mode="union")
            if "time" in keys:
                x = "time"
            elif "idx" in keys:
                x = "idx"
            else:
                x = ("idx", arange(self.n))

        xname, xdata = self.get_named_data(x)

        if label_axes:
            ax.set_xlabel(xname)
            ax.set_ylabel(yname)

        if label and isinstance(label, bool):
            label = yname

        ax.plot(xdata, ydata, label=label, **kwargs)

    def _subplots_builder(
        self,
        n_rows: int | None = 1,
        n_cols: int | None = None,
        share_xy=True,
        figsize: tuple[float, float] = (15, 5),
        idx: str = "idx",
        axs: Sequence[Axes] | None = None,
    ) -> _SubplotsBuilder:
        curves = Curves(list(self.iter_curves(idx=idx)))
        if axs is None:
            return _SubplotsBuilder.from_dims(
                curves=curves,
                nr=n_rows,
                nc=n_cols,
                sharex=share_xy,
                sharey=share_xy,
                figsize=figsize,
            )

        return _SubplotsBuilder.from_axs(curves=curves, axs=axs)

    def subplots(
        self,
        n_rows: int | None = 1,
        n_cols: int | None = None,
        axis: str | None = "scaled",
        show_axes=False,
        plot_fn: Callable[[Curve], None] | None = None,
        subtitle: str | Callable[[Curve], str] | None = None,
        share_xy=True,
        figsize: tuple[float, float] = (15, 5),
        idx: str = "idx",
        axs: Sequence[Axes] | None = None,
        hide_unused: bool = True,
    ):
        """Plot each curve in the sequence in a different subplot

        Parameters
        ----------
        figsize
            The size of the overall superfigure.

        n_rows
            Number of rows.

        n_cols
            Number of columns. By default, `n_cols` = ceil(self.n / n_rows)`. If `n_cols` is specified, and
            `n_rows` * `n_cols` < `self.n`, the curve sequence is automatically subsampled.

        axis
            Argument to `plt.axis`. By default this is 'equal' (i.e., make circles circular).

        show_axes
            Whether to show each subplot axes, i.e. border and x/y ticks, etc.

        plot_fn
            By default this just dispatches to `Curve.plot_edges(directed=False)`.

        subtitle
            A convenience argument to put a title over each subplot. If `subtitle` is a string,
            a title is constructed from the corresponding curve metadata. Otherwise, `subtitle`
            should be a function that accepts a curve and returns a string.

        share_xy
            Whether each subplot should share x/y limits.

        idx
            The index of the curve in this collection is stored in the curve metadata property
            with this name.

        axs
            Array of matplotlib axes to use for subplots. If supplied, `n_rows` and `n_cols`
            are ignored and determined by the shape of this array.

        hide_unused
            If n_axs > n_plots, hide the unused axes.

        Returns
        -------
        axes :
            `(n_rows, n_cols)` array of `matplotlib.axes.Axes` objects

        """
        if not plot_fn:

            def plot_fn(c: Curve):
                c.plot_edges(directed=False)

        if isinstance(subtitle, str):
            _key = subtitle

            def subtitle(c: Curve) -> str:
                return f"{_key} = {c[_key]}"

        b = self._subplots_builder(
            n_rows=n_rows,
            n_cols=n_cols,
            share_xy=share_xy,
            figsize=figsize,
            idx=idx,
            axs=axs,
        )
        if hide_unused and b.n_axs > b.n_plots:
            # Hide unused axes if there's e.g. 3x3 axes but only 8 curves to plot
            for i in range(b.n_plots, b.n_axs):
                b.get_ax(i).axis("off")

        for i in range(b.n_plots):
            b.axis_subplot(i=i, plot_fn=plot_fn, subtitle=subtitle, axis=axis, show_axes=show_axes)

        if b.fig is not None:  # If we created the figure
            b.fig.tight_layout()

        return b.axs

    def superimposed(
        self,
        ax: Axes | None = None,
        plot_fn: Callable[[Curve], Any] | None = None,
        color: Metadata | None = None,
        clim: tuple[float, float] | None = None,
        idx: str = "idx",
    ) -> list[Any]:
        """Plot every curve in the same axes

        Parameters
        ----------
        ax
            Matlplotlib axes to plot in. Default current axes.

        plot_fn
            Function `Curve -> matplotlib_object` to plot the curve.
            By default, this dispatches to `curve.plot`

        color
            The name of a curve metadata parameter to color by. If `plot_fn` is supplied,
            this is ignored.

        clim
            Range to scale color data to.

        idx
            The index of the curve in this collection is stored in this curve's metadata.

        Returns
        -------
        :
            List of objects returned by `plot_fn`.
        """

        ax = _get_ax(ax)

        if not plot_fn:
            if color is not None:
                cmap = plt.get_cmap("viridis")
                _cname, cdata = self.get_named_data(color)
                cnorm = plt.Normalize(*clim) if clim else plt.Normalize()
                cdata = cnorm(cdata)

                def plot_fn(c: Curve) -> Line2D:
                    return c.plot(color=cmap(cdata[c[idx]]), ax=ax)
            else:

                def plot_fn(c: Curve) -> Line2D:
                    return c.plot(ax=ax)

        out = []

        for curve in self.iter_curves(idx=idx):
            out.append(plot_fn(curve))

        return out

    def _animation_frames(self) -> Iterator[int]:
        i, step, n = 0, 1, self.n
        while True:
            yield i
            if (i, step) == (n - 1, 1):
                i, step = n - 2, -1
            elif (i, step) == (0, -1):
                i, step = 1, 1
            else:
                i += step

    def _animate(
        self,
        frames: Iterable[int] | Callable[[], int] | None = None,
        **kwargs,
    ):
        from matplotlib import animation

        kwargs.setdefault("save_count", 2 * self.n + 1)
        frames = frames or self._animation_frames()

        fig, ax = plt.subplots()
        line = self[0].plot(ax=ax)

        for c in self:
            ax.dataLim.update_from_data_xy(c.points)

        ax.axis("equal")
        ax.autoscale_view()

        def update(frame):
            curve = self[frame]
            x, y = curve.closed_points.T
            line.set_xdata(x)
            line.set_ydata(y)
            return line

        # noinspection PyTypeChecker
        return animation.FuncAnimation(
            fig=fig,
            func=update,
            frames=frames,  # type: ignore [arg-type]
            interval=30,
            **kwargs,
        )


@dataclass
class _SubplotsBuilder:
    curves: Curves
    nr: int
    nc: int
    axs: ndarray  # Should be NDArray[Axes]
    fig: Figure | None

    @property
    def n_axs(self) -> int:
        return self.nr * self.nc

    @cached_property
    def plot_idxs(self) -> ndarray:
        if self.curves.n > self.n_axs:
            step = self.curves.n // self.n_axs
            return arange(self.n_axs) * step

        return arange(self.curves.n)

    @property
    def n_plots(self) -> int:
        return min(self.curves.n, self.n_axs)

    @staticmethod
    def from_axs(curves: Curves, axs: Sequence[Axes]) -> _SubplotsBuilder:
        _axs = asanyarray(axs)
        if _axs.ndim == 1:
            nr, nc = 1, len(_axs)
        elif _axs.ndim == 2:
            nr, nc = _axs.shape
        else:
            msg = "Expected axes array to be 1 or 2 dimensional"
            raise ValueError(msg)

        return _SubplotsBuilder(curves=curves, nr=nr, nc=nc, axs=_axs, fig=None)

    @staticmethod
    def get_dims(n_curves: int, nr: int | None, nc: int | None) -> tuple[int, int]:
        if nr is not None and nc is not None:
            return nr, nc

        if nc is None and nr is not None:
            return nr, int(ceil(n_curves / nr))

        if nr is None and nc is not None:
            return int(ceil(n_curves / nc)), nc

        n = int(ceil(sqrt(n_curves)))
        return n, n

    @staticmethod
    def from_dims(
        curves: Curves,
        nr: int | None = 1,
        nc: int | None = None,
        sharex: bool = True,
        sharey: bool = True,
        figsize: tuple[float, float] = (5, 3),
    ) -> _SubplotsBuilder:
        nr_, nc_ = _SubplotsBuilder.get_dims(curves.n, nr, nc)
        fig, axs = plt.subplots(
            nrows=nr_, ncols=nc_, squeeze=False, sharex=sharex, sharey=sharey, figsize=figsize
        )

        return _SubplotsBuilder(curves=curves, nr=nr_, nc=nc_, axs=axs, fig=fig)

    def get_ax(self, i: int) -> Axes:
        return cast(Axes, self._axs_flat[i])

    @cached_property
    def _axs_flat(self) -> ndarray:
        return self.axs.reshape(-1)

    def axis_subplot(
        self,
        i: int,
        plot_fn: Callable[[Curve], Any],
        subtitle: Callable[[Curve], str] | None,
        axis: str | None,
        show_axes: bool,
    ):
        ax = self.get_ax(i)
        plt.sca(ax)  # Set current axes
        curve = self.curves.curves[self.plot_idxs[i]]
        plot_fn(curve)

        if subtitle:
            ax.set_title(subtitle(curve))

        if axis:
            ax.axis(axis)

        if not show_axes:
            # Could just do ax.axis('off') but that prevents x/y labels
            ax.spines["top"].set_visible(False)
            ax.spines["right"].set_visible(False)
            ax.spines["bottom"].set_visible(False)
            ax.spines["left"].set_visible(False)
            ax.get_xaxis().set_ticks([])
            ax.get_yaxis().set_ticks([])



"""A curve"""

from __future__ import annotations

import collections.abc
import itertools
import warnings
from collections.abc import Sequence
from functools import cached_property
from types import MappingProxyType
from typing import Any, Callable, Literal, NamedTuple, cast, overload

import numpy as np
import scipy
import shapely
import sortedcontainers
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.collections import LineCollection, PathCollection
from matplotlib.lines import Line2D
from matplotlib.quiver import Quiver
from numpy import (
    append,
    arange,
    arctan2,
    argmin,
    array,
    asanyarray,
    asarray,
    ceil,
    concatenate,
    cos,
    cross,
    cumsum,
    diff,
    gradient,
    linspace,
    ndarray,
    newaxis,
    ones,
    pi,
    repeat,
    roll,
    sign,
    sin,
    sqrt,
    stack,
    where,
    zeros,
)
from numpy.linalg import norm
from numpy.typing import ArrayLike
from typing_extensions import Self

from .util import (
    InterpType,
    _get_ax,
    _rescale,
    _VariableColorSpec,
    align_edges,
    angle_to_points,
    periodic_interpolator,
    reflection_matrix,
    rotation_matrix,
)


class Curve:
    """A discrete planar closed curve

    Parameters
    ----------
    pts
        `(n, 2)` array of vertex coordinates.

    **kwargs
        Metadata parameters in key=value format.
    """

    def __init__(
        self,
        pts: ndarray | Sequence[Sequence[float] | tuple[float, float]] | ArrayLike,
        _data: dict[str, Any] | None = None,
        **kwargs,
    ):
        # Use `asanyarray` here to allow the user to pass in whatever they want as long as it obeys
        # the numpy array protocol; in particular thinking of arrays of dual numbers for automatic
        # differentiation
        pts = asanyarray(pts)

        if pts.ndim != 2:
            msg = f"Points array must be 2-dimensional, got pts.ndims={pts.ndim}"
            raise WrongDimensions(msg)

        if pts.shape[1] != 2:
            msg = f"Points array must have 2 columns, got pts.shape={pts.shape}"
            raise WrongDimensions(msg)

        if pts.shape[0] < 3:
            msg = f"Need at least 3 points for a curve, got {pts.shape[0]}"
            raise NotEnoughPoints(msg)

        self._pts: ndarray = pts

        if kwargs and _data:
            self._data = {**_data, **kwargs}
        elif kwargs:
            self._data = kwargs
        elif _data:
            self._data = _data
        else:
            self._data = {}

    @cached_property
    def points(self) -> ndarray:
        """A `(n, 2)` array of curve vertex coordinates"""
        # Because we rely so heavily on `cached_property`s, prevent confusion due to stuff
        # like `curve.points *= 2` modifying the points array in place
        # Could just set this flag on `Curve` construction but
        #   1) The user owns the original array, not us, so don't set flag in place
        #   2) Don't want to copy the points array on construction, that's needlessly wasteful
        #   3) Assume a lot of curve construction happens from fluent chaining like
        #      `curve.scale(2).translate([3, 3]).to_length(1) where the public `points` array is
        #       never touched
        pts = self._pts.view()
        pts.flags["WRITEABLE"] = False
        return pts

    def __getitem__(self, item: str) -> Any:
        """Get curve metadata value by key name

        `curve['foo']` returns the value of the metadata parameter 'foo'.

        """
        return self._data[item]

    @property
    def data(self) -> MappingProxyType[str, Any]:
        """A read-only view of the curve's metadata"""
        return MappingProxyType(self._data)

    def with_points(self, pts: ndarray) -> Curve:
        """A curve with the newly supplied points array, but same metadata values"""
        # We can share the data dict here without a copy because it's publicly read-only
        return self.__class__(pts=pts, _data=self._data)

    def with_data(self, **kwargs) -> Curve:
        """A new curve with the same points and metadata appended with the supplied metadata

        E.g. `curve.with_data(foo=1, bar=2).with_data(baz=3)` has metadata parameters
        'foo', 'bar', and 'baz'.

        This allows without complaint overwriting previous metadata.

        Parameters
        ----------
        **kwargs
            New metadata in key=value format

        """
        return self.__class__(self._pts, _data={**self._data, **kwargs})

    def drop_data(self, *args: str) -> Curve:
        """Copy of the curve without the listed metadata parameters

        Use `curve.drop_data(*curve.data.keys())` to drop all data.
        """
        to_drop = set(args)
        data = {k: v for k, v in self._data.items() if k not in to_drop}
        return self.__class__(self._pts, _data=data)

    @property
    def n(self) -> int:
        """The number of vertices

        (or the number of edges, since this is a closed curve)

        """
        return len(self._pts)

    def __repr__(self):
        typ = self.__class__.__name__

        if self._data:
            metadata = ", ".join(f"{k}={v}" for k, v in self._data.items())
            return f"{typ}(n={self.n}, {metadata})"

        return f"{typ}(n={self.n})"

    @property
    def x(self) -> ndarray:
        """The x-component of the curve vertices"""
        return self._pts[:, 0]

    @property
    def y(self) -> ndarray:
        """The y-component of the curve vertices"""
        return self._pts[:, 1]

    @property
    def closed_points(self) -> ndarray:
        """A `(n+1, 2)` array of the vertex coordinates where the last row is equal to the first

        Curvey uses an implicitly closed representation, assuming an edge exists between the last
        and first point, i.e. in general `curve.points[0] != curve.points[-1]`. Sometimes
        it's useful to have an explicit representation.

        """
        return concatenate([self._pts, self._pts[[0]]], axis=0)

    def reverse(self, keep_first=False) -> Curve:
        """Reverse the curve orientation

        Flips between clockwise and counter-clockwise orientation

        Parameters
        ----------
        keep_first
            By default, the list of vertices is simply flipped. This changes which point is first.
            If `keep_first` is True, the points are also rolled so the first point is maintained.
        """
        pts = self._pts[::-1]

        if keep_first:
            pts = roll(pts, 1, axis=0)

        return self.with_points(pts)

    def scale(self, scale: float) -> Curve:
        """Scale vertex positions by a constant"""
        return self.with_points(scale * self._pts)

    def translate(self, offset: ndarray | Literal["center", "centroid"]) -> Curve:
        """Translate the curve

        Parameters
        ----------
        offset
            One of
                - A 2 element vector `(dx, dy)`
                - A `(n, 2)` array of `(dx, dy)` translations
                - The string 'center' or `centroid`, in which case the curve is translated so that
                  point sits on the origin.
        """
        if isinstance(offset, str):
            if offset == "center":
                offset = -self.center
            elif offset == "centroid":
                offset = -self.centroid
            else:
                raise ValueError(offset)
        else:
            offset = asarray(offset)

        return self.with_points(self._pts + offset.reshape((-1, 2)))

    def roll(self, shift: int) -> Curve:
        """Circular permutation of the vertex order

        To make vertex `i` the first vertex, use `curve.roll(-i)`.
        """
        return self.with_points(roll(self._pts, shift, axis=0))

    def rotate(self, theta: float) -> Curve:
        """Rotate the curve about the origin

        Parameters
        ----------
        theta
            Angle in radians to rotate the curve. Positive angles are counter-clockwise.
        """
        return self.transform(rotation_matrix(theta))

    def reflect(self, theta: float | Literal["x", "X", "y", "Y"]) -> Curve:
        """Reflect the curve over a line through the origin

        Parameters
        ----------
        theta
            Angle in radians of the reflection line through the origin.
            If `theta` is the string 'x' or 'y', reflect over that axis.
        """
        if isinstance(theta, str):
            if theta in ("x", "X"):
                transform = reflection_matrix(0)
            elif theta in ("y", "Y"):
                transform = reflection_matrix(pi / 2)
            else:
                msg = "Theta can only 'x', 'y', or an angle in radians"
                raise ValueError(msg)
        else:
            transform = reflection_matrix(theta)

        return self.transform(transform)

    @cached_property
    def center(self) -> ndarray:
        """The average vertex position"""
        return self._pts.mean(axis=0)

    @cached_property
    def centroid(self) -> ndarray:
        """The center of mass of the uniformly weighted polygon enclosed by the curve"""
        # https://en.wikipedia.org/wiki/Centroid#Of_a_polygon
        pts0, pts1 = self._pts, roll(self._pts, -1, axis=0)
        # xy = (x0 * y1 - x1 * y0)
        xy = (pts0 * pts1[:, ::-1] * array([[1, -1]])).sum(axis=1, keepdims=True)
        return ((pts0 + pts1) * xy).sum(axis=0) / 6 / self.signed_area

    @cached_property
    def edge(self) -> ndarray:
        """The vectors from vertex `i` to `i+1`

        See also
        --------
        [Curve.unit_edge][curvey.curve.Curve.unit_edge]
            For unit edge vectors.
        """
        pts0, pts1 = self._pts, roll(self._pts, -1, axis=0)
        return pts1 - pts0

    def to_edge_midpoints(self) -> Curve:
        """The curve whose vertices are the midpoints of this curve's edges

        Mostly just useful for plotting scalar quantities on edge midpoints.
        """
        pts = self._pts + self.edge / 2
        return self.with_points(pts)

    @cached_property
    def edge_length(self) -> ndarray:
        """Curve edge lengths

        `edge_length[i]` is the length of the edge from vertex `i` to vertex `i+1`.

        See also
        --------
        [Curve.cum_edge_length][curvey.curve.Curve.cum_edge_length]
            Cumulative egde lengths.
        """
        return norm(self.edge, axis=1)

    @cached_property
    def length(self) -> float:
        """Total arclength; the sum of edge lengths"""
        return self.edge_length.sum()

    @cached_property
    def unit_edge(self) -> ndarray:
        """The unit edge vectors from vertex `i` to `i+1`

        See also
        --------
        [Curve.tangent][curvey.curve.Curve.tangent]
            For unit tangent vectors calculated from second-order finite differences.
        """
        return self.edge / self.edge_length[:, newaxis]

    @cached_property
    def edge_normal(self) -> ndarray:
        r"""Unit edge normals

        `edge_normal[i]` is the unit vector normal to the edge from vertex `i` to `i+1`.

        Normals are computed by rotating the unit edge vectors 90 degrees counter-clockwise.

        For a counter-clockwise-oriented curve, this means that normals point inwards.

        See also
        --------
        [Curve.normal][curvey.curve.Curve.normal]
            Vertex normals calculated from 2nd order finite differences.
        """
        dx, dy = self.unit_edge.T
        # (x, y) -> (-y, x) for 90 degrees CCW rotation
        return stack([-dy, dx], axis=1)

    @cached_property
    def cum_edge_length(self) -> ndarray:
        """Cumulative edge lengths

        Simply equal to `np.cumsum(self.edge_length)`.

        `cum_edge_length` is a length `n` vector, and does not include zero,
        i.e. `curve.cum_edge_lengths[0]` is the length of the first edge, and
        `curve.cum_edge_length[-1]` == `curve.length`.

        See also
        --------
        [Curve.arclength][curvey.curve.Curve.arclength]
            Vertex arclength, like `cum_edge_length` but starts at 0.
        """
        return cumsum(self.edge_length)

    @property
    def arclength(self) -> ndarray:
        """Vertex arclengths

        `arclength` is a length `n` vector, where `arclength[i]` is the arclength
        of the `i`th vertex. `arclength[0]` is always zero. `arclength[i]` for `i>0` is equal to
        `cum_edge_length[i-1].

        See also
        --------
        [Curve.closed_arclength][curvey.curve.Curve.closed_arclength]
            Like `arclength`, but also includes `self.length` as the final element.
        """
        return append(0, self.cum_edge_length[:-1])

    @property
    def closed_arclength(self) -> ndarray:
        """Cumulative edge lengths with zero prepended

        `closed_arclength` is a length `n+1` vector, where the first element is `0`, the
        second element is the length of the first edge, and the last element is the cumulative
        length of all edges, `curve.length`.
        """
        return append(0, self.cum_edge_length)

    @cached_property
    def dual_edge_length(self) -> ndarray:
        """Vertex dual edge lengths

        `curve.dual_edge_length[i]` is the average length of the two edges incident on vertex $i$,
        i.e. $(L_{i-1, i} + L_{i, i+1})/2$  where $L_{ij}$ is the edge length between vertex
        $i$ and vertex $j$.
        """
        l_next = self.edge_length  # from vertex i to i+1
        l_prev = roll(l_next, 1)  # from i-1 to i
        return (l_prev + l_next) / 2

    @cached_property
    def tangent(self) -> ndarray:
        """Unit length tangent vectors

        `tangent[i]` is the curve unit tangent vector at vertex `i`. This is constructed from
        second order finite differences; use `Curve.unit_edge` for the exact vector from
        vertex `i` to vertex `i+1`.
        """
        # return self.edge / self.edge_length[:, newaxis]
        df_ds = self.deriv()
        df_ds /= norm(df_ds, axis=1, keepdims=True)
        return df_ds

    @cached_property
    def normal(self) -> ndarray:
        r"""Vertex unit normal vectors

        Normals are computed by rotating the unit tangents 90 degrees counter-clockwise, so that
        $\left[ T_i, N_i, 1 \right]$ forms a right-handed frame at vertex $i$ with tangent $T_i$ and
        normal $N_i$.

        For a counter-clockwise-oriented curve, this means that normals point inwards.

        See also
        --------
        [Curve.edge_normal][curvey.curve.Curve.edge_normal]
            The exact unit normals for each edge.
        """
        dx, dy = self.tangent.T
        # (x, y) -> (-y, x) for 90 degrees CCW rotation
        return stack([-dy, dx], axis=1)

    @cached_property
    def turning_angle(self) -> ndarray:
        r"""Turning angle (a.k.a exterior angle), in radians between adjacent edges

        `curve.turning_angle[i]` is the angle between the vectors $T_{i-1, i}$ and $T_{i, i+1}$
        where $T_{ij}$ is the vector from vertex $i$ to vertex $j$.

        Angles are in the range $\pm \pi$.
        """
        e_next = self.edge  # from vertex i to i+1
        e_prev = roll(e_next, 1, axis=0)  # from i-1 to i
        cos_theta = (e_prev * e_next).sum(axis=1)
        sin_theta = cross(e_prev, e_next)
        return arctan2(sin_theta, cos_theta)

    @cached_property
    def curvature(self) -> ndarray:
        r"""Length `n` vector of signed vertex curvatures

        Computed as
        $$
        \kappa_i = \frac {2 \psi_i}{L_{i-1, i} + L_{i, i+1}}
        $$
        Where $\psi_i$ is the turning angle between the two edges adjacent to vertex $i$,
        and $L_{ij}$ is the length of the edge between vertex $i$ and $j$.

        Note
        ----
        There are multiple ways to reasonably define discrete curvature. (See Table 1 in
        Vouga 2014 for a summary of their tradeoffs.) One is chosen here for
        convenience; most of the curvature flows in `curvey.flow` accept a `curvature_fn` to allow
        this choice to be overridden.
        """
        # NB self.dual_edge_length already includes the factor 2 term in the equation above
        return self.turning_angle / self.dual_edge_length

    def deriv(self, f: ndarray | None = None) -> ndarray:
        """Second order finite differences approximations of arclength-parametrized derivatives

        Derivatives are calculated by circularly padding the arclength `s` and function values
        `f(s)`, passing those to `numpy.gradient` to calculate second order finite differences,
        and then dropping the padded values.

        `f` is the function values to derivate. By default, this is the curve points, so
        `curve.deriv()` computes the curve tangent. Repeated application will compute the second
        derivative, e.g.

        ```python
        from curvey import Curve

        c = Curve.circle(n=20)
        df_ds = c.deriv()
        d2f_ds2 = c.deriv(f=df_ds)
        ```

        Parameters
        ----------
        f
            The `(n,)` or `(n, ndim)` array of function values. Defaults to `self.points`

        Returns
        -------
        deriv :
            The `(n,)` or `(n, ndim)` array of function derivature values.

        """
        if f is None:
            f = self._pts

        # Circularly pad arrays so that the derivatives of the first and last actual points are
        # calculated in the same way as the interior points
        f_periodic = concatenate([f[[-1]], f, f[[0]]], axis=0)
        s = self.cum_edge_length
        s = concatenate([[-self.edge_length[-1], 0], s])
        df_ds = gradient(f_periodic, s, axis=0)
        return df_ds[1:-1]  # Drop padded points

    @cached_property
    def signed_area(self) -> float:
        """Signed area of the polygon enclosed by the curve

        Signed area is positive if the curve is oriented counter-clockwise.

        Calculated by the [shoelace formula](https://en.wikipedia.org/wiki/Shoelace_formula).
        """
        x0, y0 = self._pts.T
        x1, y1 = roll(self._pts, -1, axis=0).T
        return 0.5 * (x0 * y1 - x1 * y0).sum()

    @cached_property
    def area(self) -> float:
        """Absolute area of the polygon enclosed by the curve"""
        return abs(self.signed_area)

    @cached_property
    def orientation(self) -> int:
        """Orientation of the curve

        Integer-valued; `+1` if curve is oriented counterclockwise, `-1` if clockwise, 0 if zero
        area

        """
        return int(sign(self.signed_area))

    def to_ccw(self) -> Curve:
        """A counterclockwise-oriented curve"""
        return self.reverse() if self.signed_area < 0 else self

    def to_cw(self) -> Curve:
        """A clockwise-oriented curve"""
        return self.reverse() if self.signed_area > 0 else self

    def to_orientation(self, orientation: int) -> Curve:
        """A curve with the specified orientation

        Parameters
        ----------
        orientation
            Must be either 1 or -1.
        """
        if orientation not in (1, -1):
            msg = f"orientation must be either 1 or -1, got {orientation}"
            raise ValueError(msg)

        return self.reverse() if self.orientation != orientation else self

    def orient_to(self, other: Curve) -> Curve:
        """A curve with the same orientation as `other`"""
        return self.to_orientation(other.orientation)

    @cached_property
    def roundness(self) -> float:
        r"""The [roundness](https://en.wikipedia.org/wiki/Roundness) of the curve

        Defined here as $P^2 / 4 \pi A$ for perimeter $P$ and area $A$.

        Equal to 1.0 for a circle and larger otherwise.
        """
        return self.length**2 / 4 / pi / self.area

    @classmethod
    def circle(cls, n: int, r: float = 1.0) -> Self:
        """Construct a regular polygon

        Parameters
        ----------
        n
            Number of vertices.

        r
            The radius.
        """
        theta = linspace(0, 2 * pi, n, endpoint=False)
        return cls(r * angle_to_points(theta))

    @classmethod
    def ellipse(cls, n: int, ra: float, rb: float) -> Self:
        """Construct an ellipse

        Parameters
        ----------
        n
            Number of vertices.

        ra
            Major radius.

        rb
            Minor radius.

        """
        theta = linspace(0, 2 * pi, n, endpoint=False)
        pts = array([[ra, rb]]) * angle_to_points(theta)
        return cls(pts)

    @classmethod
    def star(cls, n: int, r0: float, r1: float) -> Self:
        """Construct a (isotoxal) star polygon with `n` corner vertices

        Parameters
        ----------
        n
            The number of corner vertices. The returned curve has `2n` vertices.

        r0
            Radius of the even vertices.

        r1
            Radius of the odd vertices.

        """
        c = Curve.circle(n=2 * n, r=1)
        r = where(arange(c.n) % 2, r1, r0)
        return cls(r[:, newaxis] * c._pts)

    @classmethod
    def dumbbell(cls, n: int, rx: float = 2, ry: float = 2, neck: float = 0.2) -> Self:
        """Construct a dumbbell shape

        Parameters
        ----------
        n
            Number of points

        rx
            Width parameter

        ry
            Height parameter

        neck
            Height of the pinched neck
        """
        t = np.linspace(0, 1, n, endpoint=False)
        z = 2 * pi * t
        x = rx * cos(z)
        y = ry * sin(z) - (ry - neck) * sin(z) ** 3
        return cls(np.stack([x, y], axis=1))

    def drop_repeated_points(self) -> Curve:
        """Drop points that are equal to their predecessor(s)

        Repeated points result in edges with zero length and are probably bad. This will also drop
        the last point if it's equal to the first.
        """
        # NB closed_points adds the first point to the end,
        # and `diff` returns a vector one element shorter than its argument, so it all works out
        distinct = diff(self.closed_points, axis=0).any(axis=1)
        return self.with_points(self._pts[distinct])

    def interpolator(
        self,
        typ: InterpType = "cubic",
        f: ndarray | None = None,
    ) -> Callable[[ndarray], ndarray]:
        """Construct a function interpolator on curve arclength

        Parameters
        ----------
        typ
            The class of spline to use for interpolation. One of 'linear', 'cubic', or
            'pchip'.

        f
            The (n_verts,) or (n_verts, ndim) array of function values to interpolate. By default,
            this is just the vertex positions.

        Returns
        -------
        interpolator :
            A function g(s) `ndarray -> ndarray` that interpolates values of f at the arclengths s.
        """
        f = self._pts if f is None else f
        return periodic_interpolator(self.closed_arclength, f, typ=typ)

    def interpolate(self, s: ndarray, typ: InterpType = "cubic") -> Curve:
        """Construct a new curve by interpolating vertex coordinates at the supplied arclengths

        Parameters
        ----------
        s
            Arclength values to interpolate at.

        typ
            Type of interpolation, one of ('linear', 'cubic', 'pchip').
        """
        pts = self.interpolator(typ=typ, f=self._pts)(s)
        return self.with_points(pts)

    def interpolate_n(self, n: int, typ: InterpType = "cubic") -> Curve:
        """Interpolate `n` evenly spaced sampled

        Simple convenience wrapper around `Curve.interpolate`.

        Parameters
        ----------
        n
            The number of points in the new curve

        typ
            Interpolation type.
        """
        s_new = linspace(0, self.length, n, endpoint=False)
        return self.interpolate(s=s_new, typ=typ)

    def interpolate_thresh(self, thresh: float, typ: InterpType = "cubic") -> Curve:
        """Interpolate with `thresh` spacing between samples.

        Simple convenience wrapper around `Curve.interpolate`.

        Parameters
        ----------
        thresh
            Maximum arclength spacing.

        typ
            Interpolation type.
        """
        n = int(ceil(self.length / thresh))
        n = max(n, 3)
        return self.interpolate_n(n=n, typ=typ)

    @overload
    def align_to(self, target: Curve, *, return_transform: Literal[False] = False) -> Curve: ...

    @overload
    def align_to(self, target: Curve, *, return_transform: Literal[True]) -> ndarray: ...

    def align_to(self, target: Curve, *, return_transform: bool = False) -> Curve | ndarray:
        """Align to another curve by removing mean change in position and edge orientation

        Parameters
        ----------
        target
            The target curve to align to. It must have the same number of vertices as `self`. The
            edges of the other curve are assumed to be in one-to-one correspondance to the edges in
            `self`.

        return_transform
            If true, return the 3x3 transformation matrix. Otherwise, return a `Curve`

        See also
        --------
        [Curve.register_to][curvey.curve.Curve.register_to]
            Iterative closest point registration, which doesn't require corresponding vertices.

        """
        _ = self.check_same_n_vertices(target)

        if return_transform:
            return align_edges(
                self._pts, self.edge, target._pts, target.edge, return_transform=True
            )

        pts = align_edges(self._pts, self.edge, target._pts, target.edge, return_transform=False)

        return self.with_points(pts)

    def plot(self, color="black", ax: Axes | None = None, **kwargs) -> Line2D:
        """Plot the curve as a closed contour

        For more sophisticated plotting see methods `plot_points`, `plot_edges`, and `plot_vectors`.

        Parameters
        ----------
        color
            A matplotlib colorlike.

        ax
            Defaults to the current axes.

        **kwargs
            additional kwargs passed to `matplotlib.pyplot.plot`

        """
        ax = _get_ax(ax)
        (line,) = ax.plot(*self.closed_points.T, color=color, **kwargs)
        return line

    def plot_edges(
        self,
        color: ndarray | None = None,
        directed: bool = True,
        width: float | ndarray | None = None,
        scale_width: tuple[float, float] | None = None,
        ax: Axes | None = None,
        **kwargs,
    ) -> Quiver | LineCollection:
        """Plot a scalar quantity on curve edges

        Parameters
        ----------
        color
            The color to plot each edge. Defaults to curve arc length.

        directed
            If True, plot edges as arrows between vertices. Otherwise, edges are line segments.

        width
            The thickness of each edge segment, scalar or edge quantity vector.

        scale_width
            Min and max widths to scale the edge quantity to.

        ax
            The matplotlib axes to plot in. Defaults to current axes.

        **kwargs
            Aadditional kwargs passed to `plt.quiver` or `LineCollection` depending on `directed`.

        Returns
        -------
        : matplotlib.quiver.Quiver
            If `directed` is True.

        : matplotlib.collections.LineCollection
            If `directed` is False.
        """
        ax = _get_ax(ax)
        width = _rescale(width, scale_width)
        colorspec = _VariableColorSpec.parse(
            n_data=self.n, supplied=color, default_varied=self.arclength
        )

        if directed:
            if isinstance(width, (collections.abc.Sequence, ndarray)):
                msg = (
                    "`width` was supplied as a sequence but "
                    "matplotlib.pyplot.Quiver doesn't support scaling individual arrows. "
                    "Use directed=False to scale edge width."
                )
                raise ValueError(msg)

            x, y = self._pts.T
            dx, dy = self.edge.T
            return ax.quiver(
                x,
                y,
                dx,
                dy,
                *colorspec.maybe_varied,
                color=colorspec.fixed,
                angles="xy",
                scale_units="xy",
                scale=1.0,
                width=width,
                **kwargs,
            )

        # not directed
        pts = self.closed_points.reshape((-1, 1, 2))
        segments = concatenate([pts[:-1], pts[1:]], axis=1)

        if colorspec.varied is not None:
            lc = LineCollection(
                segments=segments,
                cmap="viridis",
                norm=plt.Normalize(),
                linewidths=width,
                **kwargs,
            )
            lc.set_array(colorspec.varied)
        else:
            lc = LineCollection(segments=segments, color=colorspec.fixed, linewidths=width)
        ax.add_collection(lc)

        # Adding a line collection doesn't update limits so do it here
        ax.update_datalim(self._pts)
        ax.autoscale_view()
        return lc

    def plot_vectors(
        self,
        vectors: ndarray | None = None,
        scale: ndarray | None = None,
        color=None,
        scale_length: tuple[float, float] | None = None,
        ax: Axes | None = None,
        **kwargs,
    ) -> Quiver:
        """Plot vector quantities on curve vertices

        To plot vector quantities on edges use `curve.to_edge_midpoints.plot_vectors(...)`.

        Parameters
        ----------
        vectors
            A `(n, 2)` array of vectors. Defaults to curve normals.

        scale
            A length `n` vector of length scalars to apply to the vectors.

        color
            Length `n` vector of scalar vertex quantities to color by, or a
            constant color for all edges.

        scale_length
            Limits to scale vector length to, after applying `scale`.

        ax
            The axes to plot in. Defaults to the current axes.

        **kwargs
            additional kwargs passed to `matplotlib.pyplot.quiver`
        """
        ax = _get_ax(ax)

        _vectors = self.normal if vectors is None else vectors

        if scale is not None:
            _vectors = scale.reshape(-1, 1) * _vectors

        if scale_length is not None:
            length = norm(_vectors, axis=1, keepdims=True)
            scaled_length = _rescale(length, scale_length)
            _vectors = _vectors / length * scaled_length

        colorspec = _VariableColorSpec.parse(self.n, color, default_fixed="black")

        # By default quiver doesn't include vector endpoints in x/y lim calculations
        ax.update_datalim(self._pts + _vectors)

        x, y = self._pts.T
        dx, dy = _vectors.T
        return ax.quiver(
            x,
            y,
            dx,
            dy,
            *colorspec.maybe_varied,
            color=colorspec.fixed,
            angles="xy",
            scale_units="xy",
            scale=1.0,
            **kwargs,
        )

    def plot_points(
        self,
        color: ndarray | Any | None = None,
        size: ndarray | float | None = None,
        scale_sz: tuple[float, float] | None = None,
        ax: Axes | None = None,
        **kwargs,
    ) -> PathCollection:
        """Plot a scalar quantity on curve vertices

        Parameters
        -----------
        color
            Either a matplotlib scalar colorlike or length `n` array of scalar vertex
            quantities. Defaults to `self.dual_edge_length`.

        size
            Length `n` scalar vertex quantity to size markers by, or a fixed size
            for all vertices.

        scale_sz
            Min and max sizes to scale the vertex quantity `size` to.

        ax
            Matplotlib axes to plot in. Defaults to the current axes.

        **kwargs
            additional kwargs passed to `matplotlib.pyplot.scatter`

        """
        ax = _get_ax(ax)

        if color is None:
            color = self.dual_edge_length

        size = _rescale(size, scale_sz)

        return ax.scatter(self.x, self.y, s=size, c=color, **kwargs)

    def triangulate(
        self,
        max_tri_area: float | None = None,
        min_angle: float | None = None,
        extra_params: str | None = None,
    ) -> tuple[ndarray, ndarray, ndarray]:
        """Triangulate the polygon enclosed by the curve with Shewchuck's triangulation library

        The python bindings [triangle](https://rufat.be/triangle/index.html) must be importable.
        They can be installed with `pip install triangle`.

        Parameters
        ----------
        max_tri_area: float, optional
            A global maximum triangle area constraint.

        min_angle: float, optional
            Minimum angle constraint, in degrees.

        extra_params: str, optional
            See the [API documentation](https://rufat.be/triangle/API.html).
            E.g. `extra_params='S10X' specifies a maximum number of 10 Steiner points and suppresses
            exact arithmetic.

        Returns
        -------
        points :
            The `(n, 2)` vertex coordinates of the triangulation. If `max_tri_area=None`,
            this is probably equal to `self.points`

        tris :
            `(n_tris, 3)` array of integer vertex indices.

        is_border :
            Length `n` vector of booleans, true for vertices on the border of the triangulation.

        """
        try:
            import triangle
        except ImportError as e:
            msg = "Cannot import `triangle`. Use `pip install triangle` to install."
            raise ValueError(msg) from e

        idx = arange(self.n)
        segments = stack([idx, roll(idx, -1)], axis=1)
        params = "p"  # Constrained polygon triangulation
        if max_tri_area is not None:
            params += f"a{max_tri_area: f}"
        if min_angle is not None:
            params += f"q{min_angle: f}"
        if extra_params is not None:
            params += extra_params

        d = triangle.triangulate({"vertices": self._pts, "segments": segments}, params)
        verts, tris, is_border = d["vertices"], d["triangles"], d["vertex_markers"]
        is_border = is_border.astype(bool).squeeze()
        return verts, tris, is_border

    def transform(self, transform: ndarray) -> Curve:
        """Apply a 2x2 or 3x3 transform matrix to the vertex positions"""
        pts = self._pts
        if transform.shape not in ((2, 2), (3, 3)):
            msg = f"Expected transform to have shape (2, 2) or (3, 3), got {transform.shape}"
            raise ValueError(msg)

        sz = transform.shape[0]
        if sz == 3:
            pts = concatenate([pts, ones((self.n, 1))], axis=1)
        pts = pts @ transform.T
        pts = pts[:, :2] if sz == 3 else pts
        return self.with_points(pts)

    def to_length(self, length: float = 1.0) -> Curve:
        """A new curve scaled to the supplied length"""
        return self.scale(length / self.length)

    def to_area(self, area: float = 1.0) -> Curve:
        """A new curve scaled to the supplied area"""
        return self.scale(sqrt(area / self.area))

    def subdivide(self, n: int = 1) -> Curve:
        """Create a new curve by evenly subdividing each edge

        Parameters
        ----------
        n
            Number of new points to add to each edge. For `n = 1`, new points are added at the
            edge midpoint; for `n = 2`, points are added at the one-thirds and two-thirds
            points, etc. If `n = 0`, an identical curve is returned.
        """
        if n < 0:
            msg = "n must be >= 0"
            raise ValueError(msg)

        if n == 0:
            return self

        return self.split_edges(1 + n * ones(self.n, dtype="int"))

    def split_edges(self, n: ndarray) -> Curve:
        """Sample uniformly within edge segments

        Parameters
        ----------
        n
            A integer-valued vector of length `self.n` indicating the number of points to sample
            from each edge.

            When `n[i] == 1`, simply sample vertex `i', i.e.
            `curve.split_edges(ones(curve.n, dtype='int'))` returns an identical curve.

            When `n[i] == 2`, sample at vertex `i` *and* the midpoint between vertex `i` and
            `i+1`.

            When `n[i] == 3` sample at vertex `i` and the one-third and two-thirds point, and so on.

            When `n[i] == 0`, vertex `i` is dropped from the output.

        Returns
        -------
        :
            A curve with `sum(n)` vertices.
        """

        edge_idx = repeat(arange(self.n), n)
        edge_frac = concatenate([arange(ni) / ni for ni in n])
        pts = self._pts[edge_idx] + edge_frac[:, newaxis] * self.edge[edge_idx]
        return self.with_points(pts)

    def split_long_edges(self, thresh: float) -> Curve:
        """Split edges evenly so all edge lengths are below `thresh`"""
        n_split = ceil(self.edge_length / thresh)
        return self.split_edges(n_split)

    def split_longest_edges(self, n: int) -> Curve:
        """Insert `n` new vertices by uniform edge subdivision

        Edges are split in priority of their length, so very long edges may be split into
        thirds, fourths, etc. before very short edges are split in half.
        """
        if n == 0:
            return self

        if n < 0:
            msg = "`n` must be >= 0"
            raise ValueError(msg)

        class Edge(NamedTuple):
            split_length: float
            n_subdivide: int
            idx: int
            orig_length: float

        orig_edges = (
            Edge(split_length=length, n_subdivide=1, idx=i, orig_length=length)
            for (i, length) in enumerate(self.edge_length)
        )

        # Priority queue -- break length ties by the less split edge
        queue = sortedcontainers.SortedList(orig_edges, key=lambda e: (e.length, -e.n_subdivide))

        for _ in range(n):
            edge = queue.pop()
            edge.n_subdivide += 1
            edge.split_length = edge.orig_length / edge.n_subdivide
            queue.add(edge)

        edges = sorted(queue, key=lambda e: e.idx)
        n_subdivide = array([e.n_subdivide for e in edges])
        return self.split_edges(n_subdivide)

    def collapse_shortest_edges(
        self,
        n: int | None = None,
        min_edge_length: float | None = None,
    ) -> Curve:
        """Remove vertices belonging to the shortest edges until a stopping criterion is met

        Note
        ----
        No attempt is made to prevent self-intersection.

        Parameters
        ----------
        n
            Stop after collapsing this many edges.

        min_edge_length
            Stop when the shortest edge is longer than this.

        """
        if n is None:
            n = self.n - 3

        if min_edge_length is None:
            # noinspection PyArgumentList
            min_edge_length = self.edge_length.max()

        class Vertex(NamedTuple):
            idx: int
            prev: int
            next: int
            edge_length: float  # length from vertex i to i+1

        # A doubly-linked list of vertices
        verts = {
            i: Vertex(
                idx=i,
                prev=(i - 1) % self.n,
                next=(i + 1) % self.n,
                edge_length=edge_length,
            )
            for i, edge_length in enumerate(self.edge_length)
        }

        # Priority queue by edge length
        queue = sortedcontainers.SortedSet(verts.values(), key=lambda v: -v.edge_length)

        for n_removed in itertools.count(1):
            shortest = queue.pop(-1)
            del verts[shortest.idx]

            if (n_removed == n) or (queue[-1] >= min_edge_length):
                break

            # Remove previous and next vertices so we can update them
            queue.discard(v_prev := verts[shortest.prev])
            queue.discard(v_next := verts[shortest.next])
            updated_prev_edge_length = norm(self._pts[v_next.idx] - self._pts[v_prev.idx])

            v_prev = verts[v_prev.idx] = Vertex(
                idx=v_prev.idx,
                prev=v_prev.prev,
                next=v_next.idx,
                edge_length=cast(float, updated_prev_edge_length),
            )
            v_next = verts[v_next.idx] = Vertex(
                idx=v_next.idx,
                prev=v_prev.idx,
                next=v_next.next,
                edge_length=v_next.edge_length,
            )
            queue.add(v_prev)
            queue.add(v_next)

        # Put the remaining vertices back in order
        vert_idx = array(sorted(verts.keys()))
        return self.with_points(self._pts[vert_idx])

    @cached_property
    def laplacian(self) -> scipy.sparse.dia_matrix:
        r"""The discrete Laplacian

        The Laplacian here is the graph Laplacian of a weighted graph with edge weights
        $1 / d_{i, j}$, where $d_{i, j}$ is the distance (edge length) between adjacent vertices
        $i$ and $j$.

        Returns a sparse matrix $L$ of size `(n, n)` with diagonal entries

        $$L_{i,i} = 1 / d_{i-1, i} + 1 / d_{i, i+1} $$

        and off-diagonal entries

        $$L_{i,j} = -1 / d_{i, j}$$.

        """
        return Curve._construct_laplacian(self.edge_length)

    @staticmethod
    def _construct_laplacian(edge_lengths: ndarray) -> scipy.sparse.dia_matrix:
        n = len(edge_lengths)
        l_next = 1 / edge_lengths  # l_next[i] is the inverse edge length from vertex i to i+1
        l_prev = roll(l_next, 1)  # l_prev[i] is the inverse edge length from vertex i-1 to i
        return scipy.sparse.diags(
            [
                -l_next[[-1]],  # lower corner = the single edge length (0, n-1)
                -l_next[:-1],  # lower diag = edge lengths (1, 0), (2, 1), (3, 2), ...
                l_prev + l_next,  # diagonal
                -l_next[:-1],  # upper diag = edge lengths (0, 1), (1, 2), (2, 3), ...
                -l_next[[-1]],  # upper corner = the single edge length (n-1, 0)
            ],
            offsets=(-(n - 1), -1, 0, 1, n - 1),
        ).tocsc()

    @classmethod
    def from_curvature(
        cls,
        curvature: ndarray,
        edge_lengths: ndarray,
        solve_vertices: bool = True,
        theta0: float | None = None,
        pt0: ndarray | Sequence[float] | None = None,
        dual_edge_lengths: ndarray | None = None,
        laplacian: ndarray | scipy.sparse.base.spmatrix | None = None,
    ) -> Self:
        """Construct a curve with the supplied new curvatures and edge lengths

        As explained in

        [*Robust Fairing via Conformal Curvature Flow.* Keenan Crane, Ulrich Pinkall, and
        Peter Schröder. 2014](https://www.cs.cmu.edu/~kmcrane/Projects/ConformalWillmoreFlow/paper.pdf)

        The product (curvature * edge_lengths) is integrated to obtain tangent vectors, and then
        tangent vectors are integrated to obtain vertex positions. This reconstructs the curve
        up to rotation and translation. Supply `theta0` and `pt0` to fix the orientation of the
        first edge and the location of the first point.

        This may result in an improperly closed curve. If `solve_vertices` is True, vertex
        positions are found by a linear projection to the closest closed curve, as described
        in Crane et al.

        Parameters
        ----------
        curvature
            A length `n` vector of signed vertex curvatures.

        edge_lengths
            A length `n`  vector of edge lengths. `edge_length[i]` is the distance between
            vertex `i` and `i+1`.

        theta0
            The constant of integration defining the angle of the first edge and the x-axis,
            in radians.

        pt0
            A 2-element array. The constant of integration defining the absolute position of
            the first vertex.

        solve_vertices
            If True, length discretization errors are resolved by solving
            ∆f = ▽ · T as the discrete Poisson equation Lf = b for the vertex positions f,
            as per Crane §5.2. Otherwise, vertex positions are found by simply integrating tangent
            vectors, which may result in an improperly closed contour.

        laplacian
            The `(n, n)` Laplacian array. This is constructed automatically if not supplied.

        dual_edge_lengths
            The length `n` vector of dual edge lengths. This is constructed automatically
            if not supplied.

        Examples
        --------
        Construct a circle from its expected intrinsic parameters.

        ```python
        import numpy as np
        from curvey import Curve

        n = 20
        curvatures = np.ones(n)
        edge_lengths = 2 * np.pi / n * np.ones(n)
        c = Curve.from_curvature(curvatures, edge_lengths)
        _ = c.plot_edges()
        ```

        Construct a circle from noisy parameters, using `solve_vertices` to ensure the curve
        is closed.

        ```python
        curvatures = np.random.normal(1, 0.1, n)
        edge_lengths = 2 * np.pi / n * np.random.normal(1, 0.1, n)
        c0 = Curve.from_curvature(curvatures, edge_lengths, solve_vertices=False)
        c1 = Curve.from_curvature(curvatures, edge_lengths, solve_vertices=True)
        _ = c0.plot(color='black')
        _ = c1.plot(color='red')
        ```

        """
        n = len(curvature)
        l_next = edge_lengths  # l_next[i] is the edge length from vertex i to i+1

        if dual_edge_lengths is None:
            l_prev = roll(l_next, 1)  # from i-1 to i
            # length of the edge dual to vertex i
            dual_edge_lengths = (l_next + l_prev) / 2

        # Integrate curvatures to get edge angles
        theta = zeros(n)
        if theta0 is not None:
            theta[0] = theta0
        theta[1:] = theta[0] + cumsum(curvature[1:] * dual_edge_lengths[1:])

        # Unnormalized tangent vectors from vertex i to i+1
        t_next = l_next.reshape((-1, 1)) * angle_to_points(theta)

        if solve_vertices:
            l_prev = roll(l_next, 1)  # l_prev[i] is the edge length from vertex i-1 to i
            t_prev = roll(t_next, 1, axis=0)  # Vector from vertex i-1 to i

            # `b` is the discrete divergence of the new tangent field
            b = t_prev / l_prev[:, newaxis] - t_next / l_next[:, newaxis]
            if laplacian is None:
                laplacian = Curve._construct_laplacian(edge_lengths=l_next)

            with warnings.catch_warnings():
                warnings.filterwarnings("error", category=scipy.sparse.linalg.MatrixRankWarning)

                try:
                    if scipy.sparse.issparse(laplacian):
                        pts = scipy.sparse.linalg.spsolve(laplacian, b)
                    else:
                        pts = np.linalg.solve(laplacian, b)
                # NB scipy.sparse raises numpy linalg errors
                except (np.linalg.LinAlgError, scipy.sparse.linalg.MatrixRankWarning):
                    pts = zeros((n, 2))
                    if pt0 is not None:
                        pts[0] = pt0
                    pts[1:] = pts[0] + cumsum(t_next[:-1], axis=0)

            if theta0 is not None:
                # Rotate to match requested first edge angle
                dx, dy = pts[1] - pts[0]
                theta1 = arctan2(dy, dx)
                pts @= rotation_matrix(theta1 - theta0)

            if pt0 is not None:
                pts -= pts[0] - pt0

        else:
            # Just integrate the tangent vectors
            pts = zeros((n, 2))
            if pt0 is not None:
                pts[0] = pt0
            pts[1:] = pts[0] + cumsum(t_next[:-1], axis=0)

        return cls(pts)

    def with_curvature(
        self,
        curvature: ndarray,
        solve_vertices=True,
        realign=False,
    ) -> Curve:
        """Construct a curve with the (approx.) same edge lengths and the supplied new curvatures.

        See method `Curve.from_curvature` for more details.

        Parameters
        ----------
        curvature
            A length `n` vector of signed curvatures.

        solve_vertices
            See `Curve.from_curvature`

        realign
            If True, the mean change in edge angle and vertex position is removed.
        """
        if realign:
            # Going to rotate the curve anyway, initial edge angle irrelevant
            theta0 = None
        else:
            dx, dy = self._pts[1] - self._pts[0]
            theta0 = arctan2(dy, dx)

        out = self.__class__.from_curvature(
            curvature=curvature,
            edge_lengths=self.edge_length,
            solve_vertices=solve_vertices,
            theta0=theta0,
            pt0=self._pts[0],
            dual_edge_lengths=self.dual_edge_length,
            laplacian=self.laplacian,
        )

        if realign:
            out = out.align_to(self)

        return out

    @overload
    def to_shapely(self, mode: Literal["ring"]) -> shapely.LinearRing: ...

    @overload
    def to_shapely(self, mode: Literal["edges"]) -> shapely.MultiLineString: ...

    @overload
    def to_shapely(self, mode: Literal["polygon"]) -> shapely.Polygon: ...

    @overload
    def to_shapely(self, mode: Literal["points"]) -> shapely.MultiPoint: ...

    def to_shapely(
        self,
        mode: Literal["ring", "edges", "polygon", "points"] = "ring",
    ):
        """Convenience converter to `shapely` object

        Parameters
        ----------
        mode
            Which type of `shapely` geometry to return.

              - 'ring': a `LinearRing` corresponding to the closed curve.
              - 'edges': a `MultiLineString` containing `n_edges` 2-point line segments.
              - 'polygon': a `Polygon` enclosed by the curve.
              - 'points': a `MultiPoint` containing the vertices.
        """
        if mode == "ring":
            return shapely.LinearRing(self._pts)

        if mode == "edges":
            pts0 = self._pts
            pts1 = roll(pts0, 1, axis=0)
            return shapely.MultiLineString(list(zip(pts0, pts1)))

        if mode == "polygon":
            return shapely.Polygon(self._pts)

        if mode == "points":
            return shapely.MultiPoint(self._pts)

        modes = ", ".join(("ring", "edges", "polygon", "points"))  # type: ignore [unreachable]
        msg = f"mode must be one of ({modes}), got {mode}"
        raise ValueError(msg)

    @cached_property
    def is_simple(self) -> bool:
        """False if the curve intersects or touches itself, including having repeated points

        Uses `shapely.LinearRing.is_simple`
        """
        return self.to_shapely("ring").is_simple

    def edge_intersections(self) -> ndarray:
        """An `(n_intersect, 2)` array of points where two edges cross

        This does not include two co-incident vertices or an edge coincident on
        a non-adjacent vertex.
        """
        # Use unary union to resolve all self-intersections
        mls: shapely.MultiLineString = shapely.unary_union(self.to_shapely("ring"))
        all_pts = shapely.extract_unique_points(mls)
        intersections = all_pts - self.to_shapely("points")  # Set difference

        if intersections.is_empty:
            return zeros((0, 2))

        if isinstance(intersections, shapely.Point):
            return array(intersections.coords)  # (1, 2)

        return concatenate([pt.coords for pt in intersections.geoms])

    @overload
    def register_to(
        self, target: Curve, allow_scale: bool, return_transform: Literal[False]
    ) -> Curve: ...

    @overload
    def register_to(
        self, target: Curve, allow_scale: bool, return_transform: Literal[True]
    ) -> ndarray: ...

    @overload
    def register_to(
        self, target: Curve, allow_scale: bool, return_transform: bool
    ) -> Curve | ndarray: ...

    def register_to(
        self,
        target: Curve,
        allow_scale: bool = False,
        return_transform=False,
    ) -> Curve | ndarray:
        r"""Iterative closest point registration

        Minimizes

        $$
        \sum_i \min_j d(v_i, e_j)^2
        $$

        where $d(v_i, e_j)$ is the euclidean distance btween vertices $v_i$ in `self`
        and edges $e_j$ in `target`.

        Parameters
        ----------
        target
            The `Curve` to register to/

        allow_scale
            If True, allow uniform scaling.

        return_transform
            If True, return a 3x3 transform matrix. Otherwise, return the transformed `Curve`.

        See also
        --------
        [Curve.align_to][curvey.curve.Curve.align_to]
            When `source` and `target` have the same number of vertices in 1-to-1 correspondance.
        """
        tree = shapely.STRtree(target.to_shapely("edges").geoms)

        def get_transform(params: ndarray) -> ndarray:
            """3x3 transform matrix"""
            if len(params) == 4:
                theta, dx, dy, scale_factor = params
            else:
                theta, dx, dy = params
                scale_factor = 1

            cos_theta, sin_theta = scale_factor * cos(theta), scale_factor * sin(theta)
            return array(
                [
                    [cos_theta, -sin_theta, dx],
                    [sin_theta, cos_theta, dy],
                    [0, 0, 1],
                ]
            )

        def sum_sq_dist_closest_pt(params: ndarray) -> float:
            transformed = self.transform(get_transform(params))
            (_self_idx, _other_idx), dists = tree.query_nearest(
                geometry=transformed.to_shapely("points").geoms,
                return_distance=True,
                all_matches=False,
            )
            return (dists**2).sum()

        opt = scipy.optimize.minimize(
            fun=sum_sq_dist_closest_pt,
            x0=array([0, 0, 0, 1] if allow_scale else [0, 0, 0]),
        )
        if not opt.success:
            msg = f"Optimization failed: {opt.message}"
            warnings.warn(msg, category=OptimizationFailed, stacklevel=2)

        transform = get_transform(opt.x)
        return transform if return_transform else self.transform(transform)

    def check_same_n_vertices(self, other: Curve) -> int:
        """Raises a `ValueError` if vertex counts don't match

        Otherwise, returns the common vertex count.
        """
        if self.n != other.n:
            msg = (
                "Curve pair must have the same number of vertices. "
                f"Got self.n = {self.n} and other.n = {other.n}"
            )
            raise MismatchedVertexCounts(msg)
        return self.n

    def roll_to(self, other: Curve) -> Curve:
        """Cyclicly permute points to minimize the distance between corresponding points

        `other` must have the same number of vertices as `self`
        """
        n = self.check_same_n_vertices(other)

        # (n, n) array of pairwise square distances
        dist = scipy.spatial.distance.cdist(other._pts, self._pts, "sqeuclidean")
        i0 = arange(n)

        # (n, n) cyclic permutation matrix
        #   [[0, 1, 2, 3, ..., n-1]
        #    [1, 2, 3, ..., n-1, 0]
        #    [2, 3, ..., n-1, 0, 1] ... ]
        i1 = (i0[:, newaxis] + i0[newaxis, :]) % n

        # The permutation index that minimizes sum of sq. dists
        i_min = cast(int, argmin(dist[i0, i1].sum(axis=1)))
        return self.roll(-i_min)

    def optimize_edge_lengths_to(
        self,
        other: Curve,
        interp_typ: InterpType = "cubic",
    ) -> Curve:
        """Optimize partitioning of vertex arclength positions to match edge_lengths in other

        `self` and `other` must have the same number of vertices.

        This assumes `self` and `other` have already been processed to have the same length!

        Parameters
        ----------
        other
            The curve to optimize against

        interp_typ
            Passed to `Curve.interpolator`
        """
        n = self.check_same_n_vertices(other)

        # So we don't need to refit each iteration
        interpolator = self.interpolator(typ=interp_typ)

        def _resample(ds: ndarray) -> Curve:
            arclength = append(0, cumsum(ds[:-1]))
            return self.with_points(interpolator(arclength))

        def _objective(ds: ndarray) -> float:
            resampled = _resample(ds)
            return ((other.closed_arclength - resampled.closed_arclength) ** 2).sum()

        # noinspection PyTypeChecker
        opt = scipy.optimize.minimize(
            fun=_objective,
            x0=other.edge_length,
            # Edge lengths must sum to total length
            constraints=scipy.optimize.LinearConstraint(ones(n), lb=other.length, ub=other.length),
            # Edge lengths must be positive
            bounds=scipy.optimize.Bounds(lb=0),
        )
        if not opt.success:
            msg = "Optimization failed: " + opt.message
            warnings.warn(msg, OptimizationFailed, stacklevel=2)

        return self.with_points(_resample(opt.x)._pts)


class NotEnoughPoints(Exception):
    """Raised if fewer than 3 points are passed to the `Curve` constructor"""


class WrongDimensions(Exception):
    """Raised if the points array is not 2d"""


class MismatchedVertexCounts(Exception):
    """Raised by methods expecting a pair of curves to have the same number of points"""


class OptimizationFailed(RuntimeWarning):
    """Raised by methods when optimization fails to converge"""



from __future__ import annotations

from sybil import Sybil
from sybil.parsers.markdown import PythonCodeBlockParser

# Enable sybil to collect ```python codeblocks in docstrings
pytest_collect_file = Sybil(
    parsers=[
        PythonCodeBlockParser(),
    ],
    patterns=["*.py"],
).pytest()
